# -*- coding: utf-8 -*-
#
#  Copyright 2014-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Adelia Fatykhova <adelik21979@gmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from functools import reduce
from itertools import chain, count, permutations
from logging import info
from operator import or_
from typing import Union, Iterable
from ._functions import lazy_product
from .containers import QueryContainer, QueryCGRContainer, MoleculeContainer, CGRContainer, ReactionContainer
from .periodictable import Element, DynamicElement


class BaseReactor:
    def __init__(self, reactants, products, delete_atoms):
        if isinstance(reactants, QueryContainer):
            self.__is_cgr = is_cgr = False
            e = Element
        else:
            self.__is_cgr = is_cgr = True
            e = DynamicElement

        self.__to_delete = set(reactants).difference(products) if delete_atoms else set()
        self.__elements = elements = {}
        atoms = defaultdict(dict)
        for n, atom in products.atoms():
            if atom.neighbors or atom.hybridization:
                info('neighbors and hybridization for new atoms unusable')
            atoms[n].update(charge=atom.charge, is_radical=atom.is_radical)
            elements[n] = e.from_atomic_number(atom.atomic_number)(atom.isotope)
            if n not in reactants:
                atoms[n]['xy'] = atom.xy
            if is_cgr:
                atoms[n].update(p_is_radical=atom.p_is_radical, p_charge=atom.p_charge)

        self.__atom_attrs = dict(atoms)
        self.__bond_attrs = list(products.bonds())

    def _patcher(self, structure, mapping):
        elements = self.__elements

        plane = structure._plane
        bonds = structure._bonds
        charges = structure._charges
        radicals = structure._radicals
        if self.__is_cgr:
            p_charges = structure._p_charges
            p_radicals = structure._p_radicals

        new = structure.__class__()

        to_delete = {mapping[x] for x in self.__to_delete}
        if to_delete:
            # if deleted atoms have another path to remain fragment, the path is preserved
            remain = set(mapping.values()).difference(to_delete)
            delete, global_seen = set(), set()
            for x in to_delete:
                for n in bonds[x]:
                    if n in global_seen or n in remain:
                        continue
                    seen = {n}
                    global_seen.add(n)
                    stack = [x for x in bonds[n] if x not in global_seen]
                    while stack:
                        current = stack.pop()
                        if current in remain:
                            break
                        if current in to_delete:
                            continue
                        seen.add(current)
                        global_seen.add(current)
                        stack.extend([x for x in bonds[current] if x not in global_seen])
                    else:
                        delete.update(seen)

            to_delete.update(delete)

        max_atom = max(charges) + 1
        for n, atom in self.__atom_attrs.items():
            if n in mapping:  # add matched atoms
                m = mapping[n]
                new.add_atom(elements[n].copy(), m, xy=plane[m], **atom)
            else:  # new atoms
                mapping[n] = new.add_atom(elements[n].copy(), max_atom, **atom)
                max_atom += 1

        old_atoms = set(new._atoms)
        if self.__is_cgr:
            for n, atom in structure.atoms():  # add unmatched atoms
                if n not in old_atoms and n not in to_delete:
                    new.add_atom(atom.copy(), n, charge=charges[n], is_radical=radicals[n], xy=plane[n],
                                 p_is_radical=p_radicals[n], p_charge=p_charges[n])
        else:
            for n, atom in structure.atoms():  # add unmatched atoms
                if n not in old_atoms and n not in to_delete:
                    new.add_atom(atom.copy(), n, charge=charges[n], is_radical=radicals[n], xy=plane[n])

        for n, m, bond in self.__bond_attrs:  # add patch bonds
            n = mapping[n]
            m = mapping[m]
            new.add_bond(n, m, bond.copy())

        for n, m_bond in bonds.items():
            if n in to_delete:  # atoms for removing
                continue
            to_delete.add(n)
            for m, bond in m_bond.items():
                if m in to_delete or n in old_atoms and m in old_atoms:
                    continue
                new.add_bond(n, m, bond.copy())

        # todo: calculate stereo mark based on new atom order
        return new

    def __getstate__(self):
        return {'elements': self.__elements, 'atom_attrs': self.__atom_attrs, 'bond_attrs': self.__bond_attrs,
                'is_cgr': self.__is_cgr, 'to_delete': self.__to_delete}

    def __setstate__(self, state):
        self.__elements = state['elements']
        self.__atom_attrs = state['atom_attrs']
        self.__bond_attrs = state['bond_attrs']
        self.__is_cgr = state['is_cgr']
        self.__to_delete = state['to_delete']


class CGRReactor(BaseReactor):
    """
    CGR based editor for CGRs and molecules.
    generates transformation from input CGR/molecule
    using template (CGRtools ReactionContainer).
    Template should contain one reactant and one product:

    CGRReactor calling transforms reactants to products and
    returns generator of all possible products.
    """
    def __init__(self, template: ReactionContainer, delete_atoms: bool = False):
        """
        :param template: CGRtools ReactionContainer
        :param delete_atoms: if True atoms exists in reactant but
                            not exists in product will be removed
        """
        reactants, products = template.reactants, template.products
        if not reactants or not products:
            raise ValueError('empty template')
        if any(isinstance(x, (CGRContainer, QueryCGRContainer)) for x in chain(template.reactants, template.products)):
            reactants = reduce(or_, reactants, QueryCGRContainer())
            products = reduce(or_, products, QueryCGRContainer())
        else:
            reactants = reduce(or_, reactants, QueryContainer())
            products = reduce(or_, products, QueryContainer())

        self.__pattern = reactants
        self.__meta = template.meta.copy()
        super().__init__(reactants, products, delete_atoms)

    def __call__(self, structure: Union[MoleculeContainer, CGRContainer], automorphism_filter: bool = True):
        if not isinstance(structure, (MoleculeContainer, CGRContainer)):
            raise TypeError('only Molecules and CGRs possible')

        for mapping in self.__pattern.get_mapping(structure, automorphism_filter=automorphism_filter):
            new = self._patcher(structure, mapping)
            new.meta.update(self.__meta)
            yield new

    def __getstate__(self):
        return {'pattern': self.__pattern, 'meta': self.__meta, **super().__getstate__()}

    def __setstate__(self, state):
        self.__pattern = state['pattern']
        self.__meta = state['meta']
        super().__setstate__(state)


class Reactor(BaseReactor):
    """
    CGR based reactor for molecules/queries.
    generates reaction from input queries/molecules using
    transformation template (CGRtools ReactionContainer).

    Reactor calling transforms reactants to products and
    returns generator of reaction transformations with all
    possible products
    """
    def __init__(self, template, delete_atoms=False):
        """
        :param template: CGRtools ReactionContainer
        :param delete_atoms: if True atoms exists in reactants but
                            not exists in products will be removed
        """
        reactants, products = template.reactants, template.products
        if not reactants or not products:
            raise ValueError('empty template')
        if not all(isinstance(x, (QueryContainer, MoleculeContainer)) for x in chain(products, reactants)):
            raise TypeError('only Molecules and Queries possible')

        self.__patterns = reactants = tuple(QueryContainer() | x for x in reactants)
        self.__split = len(products)

        products = reduce(or_, products, QueryContainer())
        reactants = reduce(or_, reactants)
        self.__meta = template.meta.copy()
        super().__init__(reactants, products, delete_atoms)

    def __call__(self, structures: Iterable[MoleculeContainer], automorphism_filter: bool = True):
        if any(not isinstance(structure, MoleculeContainer) for structure in structures):
            raise TypeError('only list of Molecules possible')

        structures = self.__remap(structures)
        s_nums = set(range(len(structures)))
        for chosen in permutations(s_nums, len(self.__patterns)):
            ignored = [structures[x] for x in s_nums.difference(chosen)]
            ignored_numbers = {x for x in ignored for x in x}
            chosen = [structures[x] for x in chosen]
            united_chosen = reduce(or_, chosen)
            for match in lazy_product(*(x.get_mapping(y, automorphism_filter=automorphism_filter)
                                        for x, y in zip(self.__patterns, chosen))):
                mapping = match[0]
                for m in match[1:]:
                    mapping.update(m)
                new = self._patcher(united_chosen, mapping)
                collision = set(new).intersection(ignored_numbers)
                if collision:
                    new.remap(dict(zip(collision, count(max(max(ignored_numbers), max(new.atoms_numbers)) + 1))))
                if self.__split > 1:
                    new = new.split()
                    if len(new) != self.__split:
                        info(f'expected {self.__split} molecules in reaction products, but {len(new)} formed.\n'
                             'input molecules has disconnected components')
                else:
                    new = [new]
                yield ReactionContainer(structures, new + ignored, meta=self.__meta)

    @staticmethod
    def __remap(structures):
        checked = []
        checked_atoms = set()
        for structure in structures:
            intersection = set(structure).intersection(checked_atoms)
            if intersection:
                mapping = dict(zip(intersection, count(max(max(checked_atoms), max(structure.atoms_numbers)) + 1)))
                structure = structure.remap(mapping, copy=True)
                info('some atoms in input structures had the same numbers.\n'
                     f'atoms {list(mapping)} were remapped to {list(mapping.values())}')
            checked_atoms.update(structure)
            checked.append(structure)
        return checked

    def __getstate__(self):
        return {'patterns': self.__patterns, 'meta': self.__meta, 'split': self.__split, **super().__getstate__()}

    def __setstate__(self, state):
        self.__patterns = state['patterns']
        self.__meta = state['meta']
        self.__split = state['split']
        super().__setstate__(state)


__all__ = ['CGRReactor', 'Reactor']



# -*- coding: utf-8 -*-
#
#  Copyright 2019, 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Tagir Akhmetshin <tagirshin@gmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import FrozenDict
from .element import Element
from .groups import GroupIII
from .periods import PeriodIV, PeriodV, PeriodVI, PeriodVII


class Sc(Element, PeriodIV, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 21

    @property
    def isotopes_distribution(self):
        return FrozenDict({45: 1.0, 44: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({45: 44.95591, 44: 43.959403})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()), (-3, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F')))

    @property
    def atomic_radius(self):
        return 1.84


class Y(Element, PeriodV, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 39

    @property
    def isotopes_distribution(self):
        return FrozenDict({89: 1.0, 86: 0., 90: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({89: 88.905848, 86: 85.914886, 90: 89.907152})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.12


class La(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 57

    @property
    def isotopes_distribution(self):
        return FrozenDict({138: 0.0009, 139: 0.9991})

    @property
    def isotopes_masses(self):
        return FrozenDict({138: 137.907107, 139: 138.906348})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.12  # unknown, taken radius of previous element in group


class Ce(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 58

    @property
    def isotopes_distribution(self):
        return FrozenDict({136: 0.00185, 138: 0.00251, 140: 0.8845, 142: 0.11114})

    @property
    def isotopes_masses(self):
        return FrozenDict({136: 135.90714, 138: 137.905986, 140: 139.905434, 142: 141.90924})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'O'), (1, 'O'), (1, 'O'), (1, 'O'))),
                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((2, 'O'), (2, 'O'))))

    @property
    def atomic_radius(self):
        return 2.12  # unknown, taken radius of previous element in group


class Pr(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 59

    @property
    def isotopes_distribution(self):
        return FrozenDict({141: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({141: 140.907648})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((2, 'O'), (2, 'O'))))

    @property
    def atomic_radius(self):
        return 2.47


class Nd(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 60

    @property
    def isotopes_distribution(self):
        return FrozenDict({142: 0.272, 143: 0.122, 144: 0.238, 145: 0.083, 146: 0.172, 148: 0.057, 150: 0.056})

    @property
    def isotopes_masses(self):
        return FrozenDict({142: 141.907719, 143: 142.90981, 144: 143.910083, 145: 144.912569, 146: 145.913112,
                           148: 147.916889, 150: 149.920887})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((2, 'O'),)),
                (0, False, 0, ((1, 'O'), (1, 'O'))),
                (0, False, 0, ((1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((1, 'I'), (1, 'I'))),
                (0, False, 0, ((1, 'F'), (1, 'Cl'))),
                (0, False, 0, ((1, 'F'), (1, 'Br'))),
                (0, False, 0, ((1, 'F'), (1, 'I'))),
                (0, False, 0, ((1, 'C'), (1, 'C'))),
                (0, False, 0, ((1, 'H'), (1, 'H'))))

    @property
    def atomic_radius(self):
        return 2.06


class Pm(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 61

    @property
    def isotopes_distribution(self):
        return FrozenDict({145: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({145: 144.912749})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.05


class Sm(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 62

    @property
    def isotopes_distribution(self):
        return FrozenDict({144: 0.0307, 147: 0.1499, 148: 0.1124, 149: 0.1382, 150: 0.0738, 152: 0.2675, 154: 0.2275,
                           145: 0., 153: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({144: 143.911995, 147: 146.914893, 148: 147.914818, 149: 148.91718, 150: 149.917271,
                           152: 151.919728, 154: 153.922205, 145: 144.913410, 153: 152.922097})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'O'), (1, 'O'))),
                (0, False, 0, ((1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((1, 'I'), (1, 'I'))),
                (0, False, 0, ((1, 'F'), (1, 'Cl'))),
                (0, False, 0, ((1, 'F'), (1, 'Br'))),
                (0, False, 0, ((1, 'F'), (1, 'I'))),
                (0, False, 0, ((1, 'C'), (1, 'C'))),
                (0, False, 0, ((1, 'H'), (1, 'H'))),
                (0, False, 0, ((2, 'O'),)))

    @property
    def atomic_radius(self):
        return 2.38


class Eu(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 63

    @property
    def isotopes_distribution(self):
        return FrozenDict({151: 0.4781, 153: 0.5219, 152: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({151: 150.919846, 153: 152.921226, 152: 151.921744})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'O'), (1, 'O'))),
                (0, False, 0, ((1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((1, 'I'), (1, 'I'))),
                (0, False, 0, ((1, 'F'), (1, 'Cl'))),
                (0, False, 0, ((1, 'F'), (1, 'Br'))),
                (0, False, 0, ((1, 'F'), (1, 'I'))),
                (0, False, 0, ((1, 'C'), (1, 'C'))),
                (0, False, 0, ((1, 'H'), (1, 'H'))),
                (0, False, 0, ((2, 'O'),)))

    @property
    def atomic_radius(self):
        return 2.31


class Gd(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 64

    @property
    def isotopes_distribution(self):
        return FrozenDict({152: 0.002, 154: 0.0218, 155: 0.148, 156: 0.2047, 157: 0.1565, 158: 0.2484, 160: 0.2186,
                           153: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({152: 151.919788, 154: 153.920862, 155: 154.922619, 156: 155.92212, 157: 156.923957,
                           158: 157.924101, 160: 159.927051, 153: 152.921750})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.33


class Tb(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 65

    @property
    def isotopes_distribution(self):
        return FrozenDict({159: 1.0, 160: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({159: 158.925343, 160: 159.927168})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((2, 'O'), (2, 'O'))))

    @property
    def atomic_radius(self):
        return 2.25


class Dy(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 66

    @property
    def isotopes_distribution(self):
        return FrozenDict({156: 0.0006, 158: 0.001, 160: 0.0234, 161: 0.1891, 162: 0.2551, 163: 0.249, 164: 0.2818})

    @property
    def isotopes_masses(self):
        return FrozenDict({156: 155.924278, 158: 157.924405, 160: 159.925194, 161: 160.92693, 162: 161.926795,
                           163: 162.928728, 164: 163.929171})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((2, 'O'), (2, 'O'))))

    @property
    def atomic_radius(self):
        return 2.28


class Ho(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 67

    @property
    def isotopes_distribution(self):
        return FrozenDict({165: 1.0, 166: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({165: 164.930319, 166: 165.932284})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'O'), (1, 'O'))),
                (0, False, 0, ((1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((1, 'I'), (1, 'I'))),
                (0, False, 0, ((1, 'F'), (1, 'Cl'))),
                (0, False, 0, ((1, 'F'), (1, 'Br'))),
                (0, False, 0, ((1, 'F'), (1, 'I'))),
                (0, False, 0, ((1, 'C'), (1, 'C'))),
                (0, False, 0, ((1, 'H'), (1, 'H'))),
                (0, False, 0, ((2, 'O'),)))

    @property
    def atomic_radius(self):
        return 2.26


class Er(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 68

    @property
    def isotopes_distribution(self):
        return FrozenDict({162: 0.0014, 164: 0.0161, 166: 0.3361, 167: 0.2293, 168: 0.2678, 170: 0.1493})

    @property
    def isotopes_masses(self):
        return FrozenDict({162: 161.928775, 164: 163.929197, 166: 165.93029, 167: 166.932045, 168: 167.932368,
                           170: 169.93546})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.26


class Tm(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 69

    @property
    def isotopes_distribution(self):
        return FrozenDict({169: 1.0, 170: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({169: 168.934211, 170: 169.935801})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'O'), (1, 'O'))),
                (0, False, 0, ((1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((1, 'I'), (1, 'I'))),
                (0, False, 0, ((1, 'F'), (1, 'Cl'))),
                (0, False, 0, ((1, 'F'), (1, 'Br'))),
                (0, False, 0, ((1, 'F'), (1, 'I'))),
                (0, False, 0, ((1, 'C'), (1, 'C'))),
                (0, False, 0, ((1, 'H'), (1, 'H'))),
                (0, False, 0, ((2, 'O'),)))

    @property
    def atomic_radius(self):
        return 2.22


class Yb(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 70

    @property
    def isotopes_distribution(self):
        return FrozenDict({168: 0.0013, 170: 0.0304, 171: 0.1428, 172: 0.2183, 173: 0.1613, 174: 0.3183, 176: 0.1276,
                           169: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({168: 167.933894, 170: 169.934759, 171: 170.936322, 172: 171.936378, 173: 172.938207,
                           174: 173.938858, 176: 175.942568, 169: 168.935190})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'O'), (1, 'O'))),
                (0, False, 0, ((1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((1, 'I'), (1, 'I'))),
                (0, False, 0, ((1, 'F'), (1, 'Cl'))),
                (0, False, 0, ((1, 'F'), (1, 'Br'))),
                (0, False, 0, ((1, 'F'), (1, 'I'))),
                (0, False, 0, ((1, 'C'), (1, 'C'))),
                (0, False, 0, ((1, 'H'), (1, 'H'))),
                (0, False, 0, ((2, 'O'),)))

    @property
    def atomic_radius(self):
        return 2.22


class Lu(Element, PeriodVI, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 71

    @property
    def isotopes_distribution(self):
        return FrozenDict({175: 0.9741, 176: 0.0259, 177: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({175: 174.940768, 176: 175.942682, 177: 176.943758})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.17


class Ac(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 89

    @property
    def isotopes_distribution(self):
        return FrozenDict({227: 1.0, 225: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({227: 227.027752, 225: 225.023230})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Th(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 90

    @property
    def isotopes_distribution(self):
        return FrozenDict({232: 1.0, 227: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({232: 232.03805, 227: 227.027704})

    @property
    def _common_valences(self):
        return 0, 4

    @property
    def _valences_exceptions(self):
        return ((4, False, 0, ()),
                (0, False, 0, ((1, 'Br'), (1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((1, 'I'), (1, 'I'), (1, 'I'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((1, 'I'), (1, 'I'))),
                (0, False, 0, ((1, 'H'), (1, 'H'))))

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Pa(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 91

    @property
    def isotopes_distribution(self):
        return FrozenDict({231: 1.0, 233: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({231: 231.035879, 233: 233.040247})

    @property
    def _common_valences(self):
        return 0, 4, 5

    @property
    def _valences_exceptions(self):
        return ((4, False, 0, ()),
                (0, False, 0, ((1, 'H'), (1, 'H'), (1, 'H'))),
                (0, False, 0, ((2, 'O'),)))

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class U(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 92

    @property
    def isotopes_distribution(self):
        return FrozenDict({234: 5.5e-05, 235: 0.0072, 238: 0.992745})

    @property
    def isotopes_masses(self):
        return FrozenDict({234: 234.040946, 235: 235.043923, 238: 238.050783})

    @property
    def _common_valences(self):
        return 0, 3, 4, 5, 6

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()), (4, False, 0, ()),
                (2, False, 0, ((2, 'O'), (2, 'O'))))

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Np(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 93

    @property
    def isotopes_distribution(self):
        return FrozenDict({237: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({237: 237.048173})

    @property
    def _common_valences(self):
        return 0, 2, 3, 4, 5, 6, 7

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()), (4, False, 0, ()),
                (1, False, 0, ((2, 'O'), (2, 'O'))),
                (2, False, 0, ((2, 'O'), (2, 'O'))))

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Pu(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 94

    @property
    def isotopes_distribution(self):
        return FrozenDict({239: 1.0, 242: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({239: 239.052163, 242: 242.058743})

    @property
    def _common_valences(self):
        return 0, 3, 4, 5, 6

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()), (4, False, 0, ()),
                (1, False, 0, ((2, 'O'), (2, 'O'))),
                (2, False, 0, ((2, 'O'), (2, 'O'))),
                (0, False, 0, ((2, 'Se'), )),
                (0, False, 0, ((2, 'S'),)),
                (0, False, 0, ((2, 'Te'),)),
                (0, False, 0, ((2, 'O'),)),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((1, 'I'), (1, 'I'))),
                (0, False, 0, ((1, 'H'), (1, 'H'))))

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Am(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 95

    @property
    def isotopes_distribution(self):
        return FrozenDict({241: 1.0, 243: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({241: 241.056829, 243: 243.061380})

    @property
    def _common_valences(self):
        return 0, 2, 3, 4

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Cm(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 96

    @property
    def isotopes_distribution(self):
        return FrozenDict({244: 1.0, 243: 0., 248: 0.})

    @property
    def isotopes_masses(self):
        return FrozenDict({244: 244.062753, 243: 243.061389, 248: 248.072349})

    @property
    def _common_valences(self):
        return 0, 3, 4

    @property
    def _valences_exceptions(self):
        return (0, False, 0, ((2, 'O'),)), (0, False, 0, ((1, 'H'), (1, 'H')))

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Bk(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 97

    @property
    def isotopes_distribution(self):
        return FrozenDict({249: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({249: 249.074987})

    @property
    def _common_valences(self):
        return 0, 2, 3, 4

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()), (4, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Cf(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 98

    @property
    def isotopes_distribution(self):
        return FrozenDict({249: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({249: 249.074854})

    @property
    def _common_valences(self):
        return 0, 2, 3, 4

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Es(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 99

    @property
    def isotopes_distribution(self):
        return FrozenDict({252: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({252: 252.08298})

    @property
    def _common_valences(self):
        return 0, 2, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Fm(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 100

    @property
    def isotopes_distribution(self):
        return FrozenDict({257: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({257: 257.095106})

    @property
    def _common_valences(self):
        return 0, 2, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Md(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 101

    @property
    def isotopes_distribution(self):
        return FrozenDict({258: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({258: 258.098431})

    @property
    def _common_valences(self):
        return 0, 2, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class No(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 102

    @property
    def isotopes_distribution(self):
        return FrozenDict({259: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({259: 259.10103})

    @property
    def _common_valences(self):
        return 0, 2, 3

    @property
    def _valences_exceptions(self):
        return (2, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


class Lr(Element, PeriodVII, GroupIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 103

    @property
    def isotopes_distribution(self):
        return FrozenDict({266: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({266: 266.11983})

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return (3, False, 0, ()),

    @property
    def atomic_radius(self):
        return 2.17  # unknown, taken radius of previous element in group


__all__ = ['Sc', 'Y',
           'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu',
           'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr']



# -*- coding: utf-8 -*-
#
#  Copyright 2019, 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import cached_property
from collections import defaultdict, deque
from logging import info
from typing import Dict, Optional, Set, Tuple, Union
from ..exceptions import AtomNotFound, IsChiral, NotChiral


def _pyramid_sign(n, u, v, w):
    #
    #  |   n /
    #  |   |\
    #  |   | \
    #  |  /|  \
    #  | / u---v
    #  |/___\_/___
    #        w
    #
    nx, ny, nz = n
    ux, uy, uz = u
    vx, vy, vz = v
    wx, wy, wz = w

    q1x = ux - nx
    q1y = uy - ny
    q1z = uz - nz
    q2x = vx - nx
    q2y = vy - ny
    q2z = vz - nz
    q3x = wx - nx
    q3y = wy - ny
    q3z = wz - nz

    vol = q1x * (q2y * q3z - q2z * q3y) + q1y * (q2z * q3x - q2x * q3z) + q1z * (q2x * q3y - q2y * q3x)
    if vol > 0:
        return 1
    elif vol < 0:
        return -1
    return 0


def _cis_trans_sign(n, u, v, w):
    # n      w
    #  \    /
    #   u--v
    #  /    \
    # x      x
    nx, ny = n
    ux, uy = u
    vx, vy = v
    wx, wy = w

    q1x = ux - nx
    q1y = uy - ny
    q2x = vx - ux
    q2y = vy - uy
    q3x = wx - vx
    q3y = wy - vy

    # cross vectors
    q1q2z = q1x * q2y - q1y * q2x
    q2q3z = q2x * q3y - q2y * q3x

    dot = q1q2z * q2q3z
    if dot > 0:
        return 1
    elif dot < 0:
        return -1
    return 0


def _allene_sign(n, u, v, w):
    # n    w
    # |   /
    # u--v
    nx, ny, nz = n
    ux, uy = u
    vx, vy = v
    wx, wy, wz = w

    q1x = ux - nx
    q1y = uy - ny
    q1z = -nz
    q2x = vx - ux
    q2y = vy - uy
    q3x = wx - vx
    q3y = wy - vy
    q3z = wz

    # cross vectors
    q1q2x = -q1z * q2y
    q1q2y = q1z * q2x
    q1q2z = q1x * q2y - q1y * q2x
    q2q3x = q2y * q3z
    q2q3y = -q2x * q3z
    q2q3z = q2x * q3y - q2y * q3x

    q1q2q3x = q1q2y * q2q3z - q1q2z * q2q3y
    q1q2q3y = q1q2z * q2q3x - q1q2x * q2q3z

    dot = q1q2q3x * q2x + q1q2q3y * q2y
    if dot > 0:
        return 1
    elif dot < 0:
        return -1
    return 0


class Stereo:
    __slots__ = ()

    def clean_stereo(self):
        """
        Remove stereo data
        """
        self._atoms_stereo.clear()
        self._allenes_stereo.clear()
        self._cis_trans_stereo.clear()
        self.flush_cache()

    def get_mapping(self, other, **kwargs):
        atoms_stereo = self._atoms_stereo
        allenes_stereo = self._allenes_stereo
        cis_trans_stereo = self._cis_trans_stereo
        if atoms_stereo or allenes_stereo or cis_trans_stereo:
            other_atoms_stereo = other._atoms_stereo
            other_allenes_stereo = other._allenes_stereo
            other_cis_trans_stereo = other._cis_trans_stereo
            other_translate_tetrahedron_sign = other._translate_tetrahedron_sign
            other_translate_allene_sign = other._translate_allene_sign
            other_translate_cis_trans_sign = other._translate_cis_trans_sign

            tetrahedrons = self._stereo_tetrahedrons
            cis_trans = self._stereo_cis_trans
            allenes = self._stereo_allenes

            for mapping in super().get_mapping(other, **kwargs):
                for n, s in atoms_stereo.items():
                    m = mapping[n]
                    if m not in other_atoms_stereo:  # self stereo atom not stereo in other
                        break
                    # translate stereo mark in other in order of self tetrahedron
                    if other_translate_tetrahedron_sign(m, [mapping[x] for x in tetrahedrons[n]]) != s:
                        break
                else:
                    for n, s in allenes_stereo.items():
                        m = mapping[n]
                        if m not in other_allenes_stereo:  # self stereo allene not stereo in other
                            break
                        # translate stereo mark in other in order of self allene
                        nn, nm, *_ = allenes[n]
                        if other_translate_allene_sign(m, mapping[nn], mapping[nm]) != s:
                            break
                    else:
                        for nm, s in cis_trans_stereo.items():
                            n, m = nm
                            on, om = mapping[n], mapping[m]
                            if (on, om) not in other_cis_trans_stereo:
                                if (om, on) not in other_cis_trans_stereo:
                                    break  # self stereo cis_trans not stereo in other
                                else:
                                    nn, nm, *_ = cis_trans[nm]
                                    if other_translate_cis_trans_sign(om, on, mapping[nm], mapping[nn]) != s:
                                        break
                            else:
                                nn, nm, *_ = cis_trans[nm]
                                if other_translate_cis_trans_sign(on, om, mapping[nn], mapping[nm]) != s:
                                    break
                        else:
                            yield mapping
        else:
            yield from super().get_mapping(other, **kwargs)

    @cached_property
    def _wedge_map(self):
        plane = self._plane
        atoms_stereo = self._atoms_stereo
        allenes_centers = self._stereo_allenes_centers
        atoms = self._atoms
        used = set()
        wedge = []
        for n, s in self._allenes_stereo.items():
            env = self._stereo_allenes[n]
            term = self._stereo_allenes_terminals[n]
            order = [(*env[:2], *term), (*env[1::-1], *term[::-1])]
            if env[2]:
                order.append((env[2], env[1], *term))
                order.append((env[1], env[2], *term[::-1]))
            if env[3]:
                order.append((env[3], env[0], *term[::-1]))
                order.append((env[0], env[3], *term))
            order = sorted(order, key=lambda x: (x[0] in atoms_stereo, x[0] in allenes_centers,
                                                 -atoms[x[0]].atomic_number))
            while (order[0][0], order[0][2]) in used:
                order.append(order.pop(0))
            order = order[0]
            used.add((order[2], order[0]))
            s = self._translate_allene_sign(n, *order[:2])
            v = _allene_sign((*plane[order[0]], 1), plane[order[2]], plane[order[3]], (*plane[order[1]], 0))
            if not v:
                info(f'need 2d clean. wedge stereo ambiguous for atom {{{n}}}')
            if s:
                wedge.append((order[2], order[0], v))
            else:
                wedge.append((order[2], order[0], -v))

        for n, s in atoms_stereo.items():
            order = sorted(self._stereo_tetrahedrons[n], key=lambda x: (x in atoms_stereo, x in allenes_centers,
                                                                        -atoms[x].atomic_number, atoms[x].in_ring))
            while (order[0], n) in used:
                order.append(order.pop(0))
            used.add((n, order[0]))

            s = self._translate_tetrahedron_sign(n, order)
            # need recalculation if XY changed
            if len(order) == 3:
                v = _pyramid_sign((*plane[n], 0),
                                  (*plane[order[0]], 1), (*plane[order[1]], 0), (*plane[order[2]], 0))
            else:
                v = _pyramid_sign((*plane[order[3]], 0),
                                  (*plane[order[0]], 1), (*plane[order[1]], 0), (*plane[order[2]], 0))
            if not v:
                info(f'need 2d clean. wedge stereo ambiguous for atom {{{n}}}')
            if s:
                wedge.append((n, order[0], v))
            else:
                wedge.append((n, order[0], -v))
        return tuple(wedge)

    def _translate_tetrahedron_sign(self, n, env):
        """
        Get sign of chiral tetrahedron atom for specified neighbors order

        :param n: stereo atom
        :param env: neighbors order
        """
        s = self._atoms_stereo[n]
        return self._translate_tetrahedron_sign_reversed(n , env, s)

    def _translate_tetrahedron_sign_reversed(self, n, env, s):
        order = self._stereo_tetrahedrons[n]
        if len(order) == 3:
            if len(env) == 4:  # hydrogen atom passed to env
                atoms = self._atoms
                # hydrogen always last in order
                try:
                    order = (*order, next(x for x in env if atoms[x].atomic_number == 1))  # see translate scheme
                except StopIteration:
                    raise KeyError
            elif len(env) != 3:  # pyramid or tetrahedron expected
                raise ValueError('invalid atoms list')
        elif len(env) not in (3, 4):  # pyramid or tetrahedron expected
            raise ValueError('invalid atoms list')

        translate = tuple(order.index(x) for x in env[:3])
        if _tetrahedron_translate[translate]:
            return not s
        return s

    def _translate_cis_trans_sign(self, n, m, nn, nm):
        """
        Get sign for specified opposite neighbors

        :param n: first double bonded atom
        :param m: last double bonded atom
        :param nn: neighbor of first atom
        :param nm: neighbor of last atom
        """
        try:
            s = self._cis_trans_stereo[(n, m)]
        except KeyError:
            s = self._cis_trans_stereo[(m, n)]
            n, m = m, n  # in alkenes sign not order depended
            nn, nm = nm, nn
        return self._translate_cis_trans_sign_reversed(n, m, nn, nm, s)

    def _translate_cis_trans_sign_reversed(self, n, m, nn, nm, s):
        atoms = self._atoms
        n0, n1, n2, n3 = self._stereo_cis_trans[(n, m)]
        if nn == n0:  # same start
            t0 = 0
            if nm == n1:
                t1 = 1
            elif nm == n3 or n3 is None and atoms[nm].atomic_number == 1:
                t1 = 3
            else:
                raise KeyError
        elif nn == n1:
            t0 = 1
            if nm == n0:
                t1 = 0
            elif nm == n2 or n2 is None and atoms[nm].atomic_number == 1:
                t1 = 2
            else:
                raise KeyError
        elif nn == n2 or n2 is None and atoms[nn].atomic_number == 1:
            t0 = 2
            if nm == n1:
                t1 = 1
            elif nm == n3 or n3 is None and atoms[nm].atomic_number == 1:
                t1 = 3
            else:
                raise KeyError
        elif nn == n3 or n3 is None and atoms[nn].atomic_number == 1:
            t0 = 3
            if nm == n0:
                t1 = 0
            elif nm == n2 or n2 is None and atoms[nm].atomic_number == 1:
                t1 = 2
            else:
                raise KeyError
        else:
            raise KeyError

        if _alkene_translate[(t0, t1)]:
            return not s
        return s

    def _translate_allene_sign(self, c, nn, nm):
        """
        get sign for specified opposite neighbors

        :param c: central double bonded atom
        :param nn: neighbor of first double bonded atom
        :param nm: neighbor of last double bonded atom
        """
        s = self._allenes_stereo[c]
        return self._translate_allene_sign_reversed(c, nn, nm, s)

    def _translate_allene_sign_reversed(self, c, nn, nm, s):
        atoms = self._atoms

        n0, n1, n2, n3 = self._stereo_allenes[c]
        if nn == n0:  # same start
            t0 = 0
            if nm == n1:
                t1 = 1
            elif nm == n3 or n3 is None and atoms[nm].atomic_number == 1:
                t1 = 3
            else:
                raise KeyError
        elif nn == n1:
            t0 = 1
            if nm == n0:
                t1 = 0
            elif nm == n2 or n2 is None and atoms[nm].atomic_number == 1:
                t1 = 2
            else:
                raise KeyError
        elif nn == n2 or n2 is None and atoms[nn].atomic_number == 1:
            t0 = 2
            if nm == n1:
                t1 = 1
            elif nm == n3 or n3 is None and atoms[nm].atomic_number == 1:
                t1 = 3
            else:
                raise KeyError
        elif nn == n3 or n3 is None and atoms[nn].atomic_number == 1:
            t0 = 3
            if nm == n0:
                t1 = 0
            elif nm == n2 or n2 is None and atoms[nm].atomic_number == 1:
                t1 = 2
            else:
                raise KeyError
        else:
            raise KeyError

        if _alkene_translate[(t0, t1)]:
            return not s
        return s

    @cached_property
    def _stereo_cumulenes(self) -> Dict[Tuple[int, ...], Tuple[int, int, Optional[int], Optional[int]]]:
        """
        Cumulenes which contains at least one non-hydrogen neighbor on both ends
        """
        # 5       4
        #  \     /
        #   2---3
        #  /     \
        # 1       6
        bonds = self._bonds
        atoms = self._atoms
        cumulenes = {}
        for path in self.cumulenes:
            n1, m1 = path[1], path[-2]
            nn = [x for x in bonds[path[0]] if x != n1 and atoms[x].atomic_number != 1]
            mn = [x for x in bonds[path[-1]] if x != m1 and atoms[x].atomic_number != 1]
            if nn and mn:
                sn = nn[1] if len(nn) == 2 else None
                sm = mn[1] if len(mn) == 2 else None
                cumulenes[path] = (nn[0], mn[0], sn, sm)
        return cumulenes

    @cached_property
    def _stereo_tetrahedrons(self) -> Dict[int, Union[Tuple[int, int, int], Tuple[int, int, int, int]]]:
        """
        Tetrahedrons which contains at least 3 non-hydrogen neighbors
        """
        #    2
        #    |
        # 1--K--3
        #    |
        #    4?
        atoms = self._atoms
        bonds = self._bonds
        tetrahedrons = {}
        for n in self.tetrahedrons:
            env = tuple(x for x in bonds[n] if atoms[x].atomic_number != 1)
            if len(env) in (3, 4):
                tetrahedrons[n] = env
        return tetrahedrons

    @cached_property
    def _stereo_cis_trans(self) -> Dict[Tuple[int, int], Tuple[int, int, Optional[int], Optional[int]]]:
        """
        Cis-trans bonds which contains at least one non-hydrogen neighbor on both ends
        """
        return {(n, m): env for (n, *mid, m), env in self._stereo_cumulenes.items() if not len(mid) % 2}

    @cached_property
    def _stereo_cis_trans_paths(self) -> Dict[Tuple[int, int], Tuple[int, ...]]:
        return {(path[0], path[-1]): path for path in self._stereo_cumulenes if not len(path) % 2}

    @cached_property
    def _stereo_cis_trans_terminals(self) -> Dict[int, Tuple[int, int]]:
        """
        Cis-Trans terminal atoms to cis-trans key mapping
        """
        terminals = {}
        for nm in self._stereo_cis_trans_paths:
            n, m = nm
            terminals[n] = terminals[m] = nm
        return terminals

    @cached_property
    def _stereo_allenes(self) -> Dict[int, Tuple[int, int, Optional[int], Optional[int]]]:
        """
        Allenes which contains at least one non-hydrogen neighbor on both ends
        """
        return {path[len(path) // 2]: env for path, env in self._stereo_cumulenes.items() if len(path) % 2}

    @cached_property
    def _stereo_allenes_centers(self) -> Dict[int, int]:
        """
        Allene terminal atom to center mapping
        """
        terminals = {}
        for c, (n, m) in self._stereo_allenes_terminals.items():
            terminals[n] = terminals[m] = c
        return terminals

    @cached_property
    def _stereo_allenes_terminals(self) -> Dict[int, Tuple[int, int]]:
        """
        Allene center atom to terminals mapping
        """
        return {c: (path[0], path[-1]) for c, path in self._stereo_allenes_paths.items()}

    @cached_property
    def _stereo_allenes_paths(self) -> Dict[int, Tuple[int, ...]]:
        return {path[len(path) // 2]: path for path in self._stereo_cumulenes if len(path) % 2}


class MoleculeStereo(Stereo):
    __slots__ = ()

    def add_wedge(self, n: int, m: int, mark: bool, *, clean_cache=True):
        """
        Add stereo data by wedge notation of bonds. Use it for tetrahedrons of allenes.

        :param n: number of atom from which wedge bond started
        :param m: number of atom to which wedge bond coming
        :param mark: up bond is True, down is False
        """
        if n not in self._atoms:
            raise AtomNotFound
        if n in self._atoms_stereo:
            raise IsChiral

        plane = self._plane
        if n in self._chiral_tetrahedrons:
            if m not in self._bonds[n]:
                raise AtomNotFound

            if self._atoms[m].atomic_number == 1:
                s = _pyramid_sign((*plane[m], mark), *((*plane[x], 0) for x in self._stereo_tetrahedrons[n]))
            else:
                order = [(*plane[x], mark if x == m else 0) for x in self._stereo_tetrahedrons[n]]
                if len(order) == 3:
                    s = _pyramid_sign((*plane[n], 0), *order)
                else:
                    s = _pyramid_sign(order[-1], *order[:3])
            if s:
                self._atoms_stereo[n] = s > 0
                if clean_cache:
                    self.flush_cache()
        else:
            c = self._stereo_allenes_centers.get(n)
            if c:
                if c in self._allenes_stereo:
                    raise IsChiral
                elif c not in self._chiral_allenes:
                    raise NotChiral

                order = self._stereo_allenes[c]
                t1, t2 = self._stereo_allenes_terminals[c]
                w = order.index(m)
                if w == 0:
                    m1 = order[1]
                    r = False
                elif w == 1:
                    m1 = order[0]
                    t1, t2 = t2, t1
                    r = False
                elif w == 2:
                    m1 = order[1]
                    r = True
                else:
                    m1 = order[0]
                    t1, t2 = t2, t1
                    r = True
                s = _allene_sign((*plane[m], mark), plane[t1], plane[t2], (*plane[m1], 0))
                if s:
                    self._allenes_stereo[c] = s < 0 if r else s > 0
                    if clean_cache:
                        self.flush_cache()
            else:
                # only tetrahedrons and allenes supported
                raise NotChiral

    def calculate_cis_trans_from_2d(self, *, clean_cache=True):
        """
        Calculate cis-trans stereo bonds from given 2d coordinates. Unusable for SMILES and INCHI.
        """
        cis_trans_stereo = self._cis_trans_stereo
        plane = self._plane
        flag = False
        while self._chiral_cis_trans:
            stereo = {}
            for nm in self._chiral_cis_trans:
                n, m = nm
                n1, m1, *_ = self._stereo_cis_trans[nm]
                s = _cis_trans_sign(plane[n1], plane[n], plane[m], plane[m1])
                if s:
                    stereo[nm] = s > 0
            if stereo:
                cis_trans_stereo.update(stereo)
                flag = True
                del self.__dict__['_MoleculeStereo__chiral_centers']
            else:
                break
        if flag and clean_cache:
            self.flush_cache()

    def add_atom_stereo(self, n: int, env: Tuple[int, ...], mark: bool, *, clean_cache=True):
        """
        Add stereo data for specified neighbors bypass. Use it for tetrahedrons of allenes.

        :param n: number of tetrahedron atom or central atom of allene.
        :param env: numbers of atoms with specified bypass
        :param mark: clockwise or anti bypass.

        See <https://www.daylight.com/dayhtml/doc/theory/theory.smiles.html> and <http://opensmiles.org/opensmiles.html>
        """
        if n not in self._atoms:
            raise AtomNotFound
        if n in self._atoms_stereo or n in self._allenes_stereo:
            raise IsChiral
        if not isinstance(mark, bool):
            raise TypeError('stereo mark should be bool')

        if n in self._chiral_tetrahedrons:
            self._atoms_stereo[n] = self._translate_tetrahedron_sign_reversed(n, env, mark)
            if clean_cache:
                self.flush_cache()
        elif n in self._chiral_allenes:
            self._allenes_stereo[n] = self._translate_allene_sign_reversed(n, *env, mark)
            if clean_cache:
                self.flush_cache()
        else:  # only tetrahedrons supported
            raise NotChiral

    def add_cis_trans_stereo(self, n: int, m: int, n1: int, n2: int, mark: bool, *, clean_cache=True):
        """
        Add stereo data to cis-trans double bonds (not allenes).

        n1/n=m/n2

        :param n: number of starting atom of double bonds chain (alkenes of cumulenes)
        :param m: number of ending atom of double bonds chain (alkenes of cumulenes)
        :param n1: number of neighboring atom of starting atom
        :param n2: number of neighboring atom of ending atom
        :param mark: cis or trans

        See <https://www.daylight.com/dayhtml/doc/theory/theory.smiles.html> and <http://opensmiles.org/opensmiles.html
        """
        atoms = self._atoms
        if n not in atoms or m not in atoms or n1 not in atoms or n2 not in atoms:
            raise AtomNotFound
        if not isinstance(mark, bool):
            raise TypeError('stereo mark should be bool')
        if (n, m) in self._cis_trans_stereo or (m, n) in self._cis_trans_stereo:
            raise IsChiral

        if (n, m) in self._chiral_cis_trans:
            self._cis_trans_stereo[(n, m)] = self._translate_cis_trans_sign_reversed(n, m, n1, n2, mark)
            if clean_cache:
                self.flush_cache()
        elif (m, n) in self._chiral_cis_trans:
            self._cis_trans_stereo[(m, n)] = self._translate_cis_trans_sign_reversed(m, n, n2, n1, mark)
            if clean_cache:
                self.flush_cache()
        else:
            raise NotChiral

    def _fix_stereo(self):
        if self._atoms_stereo:  # filter tetrahedrons
            stereo_tetrahedrons = self._stereo_tetrahedrons
            atoms_stereo = {k: v for k, v in self._atoms_stereo.items() if k in stereo_tetrahedrons}
            self._atoms_stereo = self_atoms_stereo = {}
        else:
            atoms_stereo = {}

        if self._allenes_stereo:  # filter allenes
            stereo_allenes = self._stereo_allenes
            allenes_stereo = {k: v for k, v in self._allenes_stereo.items() if k in stereo_allenes}
            self._allenes_stereo = self_allenes_stereo = {}
        else:
            allenes_stereo = {}

        if self._cis_trans_stereo:  # filter cis-trans
            stereo_cis_trans = self._stereo_cis_trans
            cis_trans_stereo = {k: v for k, v in self._cis_trans_stereo.items() if k in stereo_cis_trans}
            self._cis_trans_stereo = self_stereo_cis_trans = {}
        else:
            cis_trans_stereo = {}

        old_stereo = len(atoms_stereo) + len(allenes_stereo) + len(cis_trans_stereo)
        while old_stereo:
            chiral_tetrahedrons = self._chiral_tetrahedrons
            chiral_allenes = self._chiral_allenes
            chiral_cis_trans = self._chiral_cis_trans

            tmp = {}
            for n, s in atoms_stereo.items():
                if n in chiral_tetrahedrons:
                    self_atoms_stereo[n] = s
                else:
                    tmp[n] = s
            atoms_stereo = tmp

            tmp = {}
            for n, s in allenes_stereo.items():
                if n in chiral_allenes:
                    self_allenes_stereo[n] = s
                else:
                    tmp[n] = s
            allenes_stereo = tmp

            tmp = {}
            for n, s in cis_trans_stereo.items():
                if n in chiral_cis_trans:
                    self_stereo_cis_trans[n] = s
                else:
                    tmp[n] = s
            cis_trans_stereo = tmp

            fail_stereo = len(atoms_stereo) + len(allenes_stereo) + len(cis_trans_stereo)
            if fail_stereo == old_stereo:
                break
            old_stereo = fail_stereo
            # flush cache
            del self.__dict__['_MoleculeStereo__chiral_centers']

    @property
    def _chiral_tetrahedrons(self) -> Set[int]:
        return self.__chiral_centers[0]

    @property
    def _chiral_cis_trans(self) -> Set[Tuple[int, int]]:
        return self.__chiral_centers[1]

    @property
    def _chiral_allenes(self) -> Set[int]:
        return self.__chiral_centers[2]

    @cached_property
    def _stereo_axises(self) -> Tuple[Tuple[Tuple[int, ...], ...], Tuple[Tuple[int, ...], ...]]:
        """
        Get all stereogenic axises in rings with attached cumulenes. Stereogenic axises has only stereogenic atoms.
        """
        morgan = self.atoms_order
        bonds = self._bonds
        stereo_tetrahedrons = self._stereo_tetrahedrons
        cumulenes_terminals = {}
        for n, *_, m in self._stereo_cumulenes:
            cumulenes_terminals[n] = m
            cumulenes_terminals[m] = n

        out = []
        env = []
        axises = set()
        for c in self.connected_rings_cumulenes:
            out_c = []
            env_c = []
            adj = {n: {m: 1 for m, b in bonds[n].items() if m in c} for n in c}
            w_atoms = {n: morgan[n] for n in c}
            for mapping in self._get_automorphism_mapping(w_atoms, adj, w_atoms):
                sym = {k for k, v in mapping.items() if k != v}
                if not sym:
                    continue
                # get self-matched atoms connected with automorphic atoms
                ax = {k for k in mapping.keys() - sym if not sym.isdisjoint(bonds[k])}
                # add terminal stereogenic cumulenes
                c = 0
                tmp = set()
                for n in ax:
                    if n in stereo_tetrahedrons:
                        c += 1
                        tmp.add(n)
                    elif n in cumulenes_terminals:
                        tmp.add(n)
                        m = cumulenes_terminals[n]
                        if m not in ax:
                            tmp.add(m)
                            c += 1
                if c < 2:
                    continue
                ax = frozenset(tmp)
                if ax in axises:
                    continue
                axises.add(ax)
                if len(ax) > 2:  # get order of atoms
                    start = next(iter(ax))
                    # prepare distances
                    dist = {start: 0}
                    queue = deque([(start, 1)])
                    while queue:
                        n, d = queue.popleft()
                        d1 = d + 1
                        for m in adj[n].keys() - dist.keys():
                            queue.append((m, d1))
                            dist[m] = d
                    # get paths
                    seen = {start}
                    path = None
                    cut = {start}
                    stack = deque()
                    for n in adj[start]:
                        if mapping[n] not in cut:
                            stack.append((n, 0))
                            cut.add(n)
                    full_path = None
                    while stack:
                        n, d = stack.pop()
                        if not d:  # lets start
                            if path:
                                full_path = path[::-1]
                            path = [start]
                            d = 1
                        elif dist[n] != d:
                            continue
                        elif len(path) > d:
                            path = path[:d]
                        path.append(n)
                        if n in ax:
                            seen.add(n)
                            if seen == ax:
                                break
                        d += 1
                        for m in adj[n]:
                            if m in cut:
                                if m not in path:
                                    stack.append((m, d))
                            elif mapping[m] not in cut:
                                cut.add(m)
                                if m not in path:
                                    stack.append((m, d))

                    if full_path:
                        path = full_path + path[1:]
                    ax = tuple(n for n in path if n in ax)
                else:
                    ax = tuple(ax)
                out_c.append(ax)
                env_c.append(tuple(n for n in sym if n in bonds[ax[0]] or n in bonds[ax[-1]]))
            for ax, e in sorted(zip(out_c, env_c), key=lambda x: len(x[0]), reverse=True):
                out.append(ax)
                env.append(e)
        return tuple(out), tuple(env)

    @cached_property
    def __stereo_axises(self):
        bonds = self._bonds
        tetrahedrons = self._stereo_tetrahedrons
        cumulenes = self._stereo_cumulenes

        cumulenes_terminals = {}
        for path in cumulenes:
            cumulenes_terminals[path[0]] = cumulenes_terminals[path[-1]] = path
        axises = []
        for ax, env in zip(*self._stereo_axises):
            ax_t, ax_a, ax_c = set(), set(), set()
            checks = {}
            axises.append((ax_t, ax_a, ax_c, checks))
            for n in ax:
                if n in tetrahedrons:
                    ax_t.add(n)
                else:
                    path = cumulenes_terminals[n]
                    if len(path) % 2:
                        ax_a.add(path)
                    else:
                        ax_c.add(path)
            for n in (ax[0], ax[-1]):
                if n in tetrahedrons:
                    ngb = tuple(m for m in tetrahedrons[n] if m not in env)
                else:
                    path = cumulenes_terminals[n]
                    ngb = tuple(m for m in bonds[n] if m not in path)
                if len(ngb) == 2:  # only this atoms should be checked
                    checks[n] = ngb
        return axises

    @cached_property
    def __chiral_centers(self):
        atoms_stereo = self._atoms_stereo
        cis_trans_stereo = self._cis_trans_stereo
        allenes_stereo = self._allenes_stereo

        morgan = self.atoms_order
        tetrahedrons = self._stereo_tetrahedrons.copy()
        cumulenes = self._stereo_cumulenes.copy()
        axises = self.__stereo_axises

        morgan_update = {}
        while True:
            # tetrahedron is chiral if all its neighbors are unique.
            chiral_t = {n for n, env in tetrahedrons.items() if len({morgan[x] for x in env}) == len(env)}
            # double bond is chiral if neighbors of each terminal atom is unique.
            chiral_c = set()
            chiral_a = set()
            for path, (n1, m1, n2, m2) in cumulenes.items():
                if morgan[n1] != morgan.get(n2, 0) and morgan[m1] != morgan.get(m2, 0):
                    if len(path) % 2:
                        chiral_a.add(path)
                    else:
                        chiral_c.add(path)
            # axis with 2 terminal chiral atoms is chiral
            for ax in axises:
                ax_t, ax_a, ax_c, check = ax
                if not chiral_t.isdisjoint(ax_t) or not ax_a.isdisjoint(chiral_a) or not ax_c.isdisjoint(chiral_c):
                    continue  # self chiral centers can't be in axises
                elif check and any(morgan[n] == morgan[m] for n, m in check.values()):  # need additional check
                    continue  # not chiral
                chiral_t.update(ax_t)
                chiral_a.update(ax_a)
                chiral_c.update(ax_c)

            # separate equal constitutionally but unique by stereo type chiral centers
            # need for searching depended chiral centers
            if atoms_stereo:
                grouped_stereo = defaultdict(list)
                for n in chiral_t:
                    if n in atoms_stereo:
                        grouped_stereo[morgan[n]].append(n)  # collect equal stereo atoms
                for group in grouped_stereo.values():
                    if not len(group) % 2:  # only even number of equal stereo atoms give new stereo center
                        s = [n for n in group
                             if self._translate_tetrahedron_sign(n, sorted(tetrahedrons[n], key=morgan.get))]
                        if 0 < len(s) < len(group):  # RS pair required
                            for n in s:
                                morgan_update[n] = -morgan[n]
                    for n in group:  # remove seen stereo atoms
                        del tetrahedrons[n]
                        chiral_t.discard(n)

            if cis_trans_stereo:
                grouped_stereo = defaultdict(list)
                for path in chiral_c:
                    n, *_, m = path
                    mn, mm = morgan[n], morgan[m]
                    if (n, m) in cis_trans_stereo or (m, n) in cis_trans_stereo:
                        if mn <= mm:
                            grouped_stereo[mn].append((n, m, cumulenes[path], path))
                        else:
                            grouped_stereo[mm].append((m, n, cumulenes[path], path))
                for group in grouped_stereo.values():
                    if not len(group) % 2:  # only even number of equal stereo bonds give new stereo center
                        s = []
                        for n, m, (n1, m1, n2, m2), _ in group:
                            if n2 is None:
                                a = n1
                            else:
                                a = min(n1, n2, key=morgan.get)
                            if m2 is None:
                                b = m1
                            else:
                                b = min(m1, m2, key=morgan.get)
                            if self._translate_cis_trans_sign(n, m, a, b):
                                s.append(n)
                        if 0 < len(s) < len(group):  # RS pair required
                            for n in s:
                                morgan_update[n] = -morgan[n]
                    for *_, path in group:  # remove seen stereo atoms
                        del cumulenes[path]
                        chiral_c.discard(path)

            if allenes_stereo:
                grouped_stereo = defaultdict(list)
                for path in chiral_a:
                    c = path[len(path) // 2]
                    grouped_stereo[morgan[c]].append((c, cumulenes[path], path))
                for group in grouped_stereo.values():
                    if not len(group) % 2:  # only even number of equal stereo bonds give new stereo center
                        s = []
                        for c, (n1, m1, n2, m2), _ in group:
                            if n2 is None:
                                a = n1
                            else:
                                a = min(n1, n2, key=morgan.get)
                            if m2 is None:
                                b = m1
                            else:
                                b = min(m1, m2, key=morgan.get)
                            if self._translate_allene_sign(c, a, b):
                                s.append(c)
                        if 0 < len(s) < len(group):  # RS pair required
                            for c in s:
                                morgan_update[c] = -morgan[c]
                    for *_, path in group:  # remove seen stereo atoms
                        del cumulenes[path]
                        chiral_a.discard(path)

            tmp = []
            for ax in axises:
                ax_t, ax_a, ax_c, check = ax
                # remove fully labeled axises
                if ax_t.issubset(tetrahedrons) and ax_a.issubset(cumulenes) and ax_c.issubset(cumulenes):
                    tmp.append(ax)
            axises = tmp

            if morgan_update:
                morgan = self._morgan({**morgan, **morgan_update})
                morgan_update = {}
            else:
                return chiral_t, {(n, m) for n, *_, m in chiral_c}, {path[len(path) // 2] for path in chiral_a}


class QueryStereo(Stereo):  # todo: implement add_wedge
    __slots__ = ()

    def add_wedge(self, n: int, m: int, mark: bool, *, clean_cache=True):
        raise NotImplementedError


# 1  2
#  \ |
#   \|
#    n---3
#   /
#  /
# 0
_tetrahedron_translate = {(0, 1, 2): False, (1, 2, 0): False, (2, 0, 1): False,
                          (0, 2, 1): True, (1, 0, 2): True, (2, 1, 0): True,
                          (0, 3, 1): False, (3, 1, 0): False, (1, 0, 3): False,
                          (0, 1, 3): True, (1, 3, 0): True, (3, 0, 1): True,
                          (0, 2, 3): False, (2, 3, 0): False, (3, 0, 2): False,
                          (0, 3, 2): True, (3, 2, 0): True, (2, 0, 3): True,
                          (1, 3, 2): False, (3, 2, 1): False, (2, 1, 3): False,
                          (1, 2, 3): True, (2, 3, 1): True, (3, 1, 2): True}
# 2       1
#  \     /
#   n---m
#  /     \
# 0       3
_alkene_translate = {(0, 1): False, (1, 0): False, (0, 3): True, (3, 0): True,
                     (2, 3): False, (3, 2): False, (2, 1): True, (1, 2): True}


__all__ = ['MoleculeStereo', 'QueryStereo']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from typing import List, Tuple, Union, Dict
from . import cgr, molecule  # cyclic imports resolve
from .bonds import Bond
from .common import Graph
from ..algorithms.calculate2d import Calculate2DMolecule
from ..algorithms.components import StructureComponents
from ..algorithms.depict import DepictQuery
from ..algorithms.smiles import QuerySmiles
from ..algorithms.stereo import QueryStereo
from ..periodictable import Element, QueryElement, AnyElement


class QueryContainer(QueryStereo, Graph, QuerySmiles, StructureComponents, DepictQuery, Calculate2DMolecule):
    __slots__ = ('_neighbors', '_hybridizations', '_atoms_stereo', '_cis_trans_stereo', '_allenes_stereo')

    def __init__(self):
        self._neighbors: Dict[int, Tuple[int, ...]] = {}
        self._hybridizations: Dict[int, Tuple[int, ...]] = {}
        self._atoms_stereo: Dict[int, bool] = {}
        self._allenes_stereo: Dict[int, bool] = {}
        self._cis_trans_stereo: Dict[Tuple[int, int], bool] = {}

        super().__init__()

    def add_atom(self, atom: Union[QueryElement, AnyElement, Element, int, str], *args,
                 neighbors: Union[int, List[int], Tuple[int, ...], None] = None,
                 hybridization: Union[int, List[int], Tuple[int, ...], None] = None, **kwargs):
        neighbors = self._validate_neighbors(neighbors)
        hybridization = self._validate_hybridization(hybridization)

        if not isinstance(atom, (QueryElement, AnyElement)):
            if isinstance(atom, Element):
                atom = QueryElement.from_atomic_number(atom.atomic_number)(atom.isotope)
            elif isinstance(atom, str):
                atom = QueryElement.from_symbol(atom)()
            elif isinstance(atom, int):
                atom = QueryElement.from_atomic_number(atom)()
            else:
                raise TypeError('QueryElement object expected')

        _map = super().add_atom(atom, *args, **kwargs)
        self._neighbors[_map] = neighbors
        self._hybridizations[_map] = hybridization
        return _map

    def add_bond(self, n, m, bond: Union[Bond, int]):
        if not isinstance(bond, Bond):
            bond = Bond(bond)

        sct = self._stereo_cis_trans_paths  # save
        sa = self._stereo_allenes_paths

        super().add_bond(n, m, bond)
        # remove stereo marks on bonded atoms and all its bonds
        if n in self._atoms_stereo:
            del self._atoms_stereo[n]
        if m in self._atoms_stereo:
            del self._atoms_stereo[m]
        if self._cis_trans_stereo:
            for nm, path in sct.items():
                if (n in path or m in path) and nm in self._cis_trans_stereo:
                    del self._cis_trans_stereo[nm]
        if self._allenes_stereo:
            for c, path in sa.items():
                if (n in path or m in path) and c in self._allenes_stereo:
                    del self._allenes_stereo[c]

    def delete_atom(self, n):
        bonds = set(self._bonds[n])  # save
        sct = self._stereo_cis_trans_paths
        sa = self._stereo_allenes_paths

        super().delete_atom(n)

        del self._neighbors[n]
        del self._hybridizations[n]

        sas = self._atoms_stereo
        if n in sas:
            del sas[n]
        for m in bonds:
            if m in sas:
                del sas[m]
        if self._cis_trans_stereo:
            for nm, path in sct.items():
                if not bonds.isdisjoint(path) and nm in self._cis_trans_stereo:
                    del self._cis_trans_stereo[nm]
        if self._allenes_stereo:
            for c, path in sa.items():
                if not bonds.isdisjoint(path) and c in self._allenes_stereo:
                    del self._allenes_stereo[c]

    def delete_bond(self, n, m):
        sct = self._stereo_cis_trans_paths  # save
        sa = self._stereo_allenes_paths

        super().delete_bond(n, m)

        if n in self._atoms_stereo:
            del self._atoms_stereo[n]
        if m in self._atoms_stereo:
            del self._atoms_stereo[m]
        if self._cis_trans_stereo:
            for nm, path in sct.items():
                if (n in path or m in path) and nm in self._cis_trans_stereo:
                    del self._cis_trans_stereo[nm]
        if self._allenes_stereo:
            for c, path in sa.items():
                if (n in path or m in path) and c in self._allenes_stereo:
                    del self._allenes_stereo[c]

    def remap(self, mapping, *, copy=False) -> 'QueryContainer':
        h = super().remap(mapping, copy=copy)
        mg = mapping.get
        sn = self._neighbors

        if copy:
            hn = h._neighbors
            hh = h._hybridizations
            has = h._atoms_stereo
            hal = h._allenes_stereo
            hcs = h._cis_trans_stereo
        else:
            hn = {}
            hh = {}
            has = {}
            hal = {}
            hcs = {}

        for n, hyb in self._hybridizations.items():
            m = mg(n, n)
            hn[m] = sn[n]
            hh[m] = hyb

        for n, stereo in self._atoms_stereo.items():
            has[mg(n, n)] = stereo
        for n, stereo in self._allenes_stereo.items():
            hal[mg(n, n)] = stereo
        for (n, m), stereo in self._cis_trans_stereo.items():
            hcs[(mg(n, n), mg(m, m))] = stereo

        if copy:
            return h

        self._neighbors = hn
        self._hybridizations = hh
        self._atoms_stereo = has
        self._allenes_stereo = hal
        self._cis_trans_stereo = hcs
        return self

    def copy(self, **kwargs) -> 'QueryContainer':
        copy = super().copy(**kwargs)
        copy._neighbors = self._neighbors.copy()
        copy._hybridizations = self._hybridizations.copy()
        copy._atoms_stereo = self._atoms_stereo.copy()
        copy._allenes_stereo = self._allenes_stereo.copy()
        copy._cis_trans_stereo = self._cis_trans_stereo.copy()
        return copy

    def substructure(self, atoms, **kwargs) -> 'QueryContainer':
        """
        create substructure containing atoms from atoms list

        :param atoms: list of atoms numbers of substructure
        :param meta: if True metadata will be copied to substructure
        """
        sub, atoms = super().substructure(atoms, self.__class__, **kwargs)
        sa = self._atoms
        sb = self._bonds
        sn = self._neighbors
        sh = self._hybridizations

        sub._neighbors = {n: sn[n] for n in atoms}
        sub._hybridizations = {n: sh[n] for n in atoms}

        lost = {n for n, a in sa.items() if a.atomic_number != 1} - set(atoms)  # atoms not in substructure
        not_skin = {n for n in atoms if lost.isdisjoint(sb[n])}
        sub._atoms_stereo = {n: s for n, s in self._atoms_stereo.items() if n in not_skin}
        sub._allenes_stereo = {n: s for n, s in self._allenes_stereo.items() if
                               not_skin.issuperset(self._stereo_allenes_paths[n]) and not_skin.issuperset(
                                       x for x in self._stereo_allenes[n] if x)}
        sub._cis_trans_stereo = {nm: s for nm, s in self._cis_trans_stereo.items() if
                                 not_skin.issuperset(self._stereo_cis_trans_paths[nm]) and not_skin.issuperset(
                                         x for x in self._stereo_cis_trans[nm] if x)}

        sub._atoms = ca = {}
        for n in atoms:
            atom = sa[n].copy()
            ca[n] = atom
            atom._attach_to_graph(sub, n)
        return sub

    def union(self, other, **kwargs) -> 'QueryContainer':
        if isinstance(other, (QueryContainer, molecule.MoleculeContainer)):
            u, other = super().union(other, **kwargs)
            if isinstance(other, QueryContainer):
                u._neighbors.update(other._neighbors)
                u._hybridizations.update(other._hybridizations)

                ua = u._atoms
                for n, atom in other._atoms.items():
                    atom = atom.copy()
                    ua[n] = atom
                    atom._attach_to_graph(u, n)
            else:
                un = u._neighbors
                uh = u._hybridizations
                oh = other._hybridizations
                for n, m_bond in other._bonds.items():
                    un[n] = (len(m_bond),)
                    uh[n] = (oh[n],)

                ua = u._atoms
                for n, atom in other._atoms.items():
                    atom = QueryElement.from_atomic_number(atom.atomic_number)(atom.isotope)
                    ua[n] = atom
                    atom._attach_to_graph(u, n)

            ub = u._bonds
            for n, m_bond in other._bonds.items():
                ub[n] = ubn = {}
                for m, bond in m_bond.items():
                    if m in ub:  # bond partially exists. need back-connection.
                        ubn[m] = ub[m][n]
                    else:
                        ubn[m] = bond.copy()

            u._atoms_stereo.update(other._atoms_stereo)
            u._allenes_stereo.update(other._allenes_stereo)
            u._cis_trans_stereo.update(other._cis_trans_stereo)
            return u
        elif isinstance(other, cgr.CGRContainer):
            raise TypeError('QueryContainer and CGRContainer unite impossible')
        elif isinstance(other, Graph):
            return other.union(self, **kwargs)
        else:
            raise TypeError('Graph expected')

    def get_mapping(self, other: Union['QueryContainer', 'molecule.MoleculeContainer'], **kwargs):
        if isinstance(other, (QueryContainer, molecule.MoleculeContainer)):
            return super().get_mapping(other, **kwargs)
        raise TypeError('MoleculeContainer or QueryContainer expected')

    def get_mcs_mapping(self, other: Union['QueryContainer', 'molecule.MoleculeContainer'], **kwargs):
        if isinstance(other, (QueryContainer, molecule.MoleculeContainer)):
            return super().get_mcs_mapping(other, **kwargs)
        raise TypeError('MoleculeContainer or QueryContainer expected')

    @staticmethod
    def _validate_neighbors(neighbors):
        if neighbors is None:
            neighbors = ()
        elif isinstance(neighbors, int):
            if neighbors < 0 or neighbors > 14:
                raise ValueError('neighbors should be in range [0, 14]')
            neighbors = (neighbors,)
        elif isinstance(neighbors, (tuple, list)):
            if not all(isinstance(n, int) for n in neighbors):
                raise TypeError('neighbors should be list or tuple of ints')
            if any(n < 0 or n > 14 for n in neighbors):
                raise ValueError('neighbors should be in range [0, 14]')
            if len(set(neighbors)) != len(neighbors):
                raise ValueError('neighbors should be unique')
            neighbors = tuple(sorted(neighbors))
        else:
            raise TypeError('neighbors should be int or list or tuple of ints')
        return neighbors

    @staticmethod
    def _validate_hybridization(hybridization):
        if hybridization is None:
            hybridization = ()
        elif isinstance(hybridization, int):
            if hybridization < 1 or hybridization > 4:
                raise ValueError('hybridization should be in range [1, 4]')
            hybridization = (hybridization,)
        elif isinstance(hybridization, (tuple, list)):
            if not all(isinstance(h, int) for h in hybridization):
                raise TypeError('hybridizations should be list or tuple of ints')
            if any(h < 1 or h > 4 for h in hybridization):
                raise ValueError('hybridizations should be in range [1, 4]')
            if len(set(hybridization)) != len(hybridization):
                raise ValueError('hybridizations should be unique')
            hybridization = tuple(sorted(hybridization))
        else:
            raise TypeError('hybridization should be int or list or tuple of ints')
        return hybridization

    def __getstate__(self):
        return {'atoms_stereo': self._atoms_stereo, 'allenes_stereo': self._allenes_stereo,
                'cis_trans_stereo': self._cis_trans_stereo, 'neighbors': self._neighbors,
                'hybridizations': self._hybridizations, **super().__getstate__()}

    def __setstate__(self, state):
        if 'node' in state:  # 3.1 compatibility.
            state['atoms'] = {n: a.atom for n, a in state['node'].items()}
            state['charges'] = {n: a.charge for n, a in state['node'].items()}
            state['radicals'] = {n: a.is_radical for n, a in state['node'].items()}
            state['neighbors'] = {n: a.neighbors for n, a in state['node'].items()}
            state['hybridizations'] = {n: a.hybridization for n, a in state['node'].items()}
            state['parsed_mapping'] = {}
            state['bonds'] = bonds = {}
            for n, m_bond in state['adj'].items():
                bonds[n] = bn = {}
                for m, bond in m_bond.items():
                    if m in bonds:
                        bn[m] = bonds[m][n]
                    else:
                        bn[m] = Bond(bond.order)

            state['plane'] = {n: a.xy for n, a in state['node'].items()}
            state['atoms_stereo'] = {}
            state['allenes_stereo'] = {}
            state['cis_trans_stereo'] = {}
        elif 'allenes_stereo' not in state:  # <4.0.22
            state['atoms_stereo'] = {}  # flush existing stereo if exists.
            state['allenes_stereo'] = {}
            state['cis_trans_stereo'] = {}

        super().__setstate__(state)
        self._atoms_stereo = state['atoms_stereo']
        self._allenes_stereo = state['allenes_stereo']
        self._cis_trans_stereo = state['cis_trans_stereo']
        self._neighbors = state['neighbors']
        self._hybridizations = state['hybridizations']


__all__ = ['QueryContainer']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import cached_property
from collections import Counter
from itertools import groupby
from logging import warning
from operator import itemgetter
from sys import version_info
from typing import Dict


if version_info[1] >= 8:
    tuple_hash = hash
else:
    def tuple_hash(v):
        """
        Python 3.8 hash for tuples implemented on python.
        Working only for nonnested tuples.
        """
        acc = 0x27D4EB2F165667C5
        for el in v:
            lane = hash(el)
            if lane == -1:
                return -1
            elif lane < 0:
                lane += 0x10000000000000000  # to unsigned

            acc += lane * 0xC2B2AE3D27D4EB4F
            acc %= 0x10000000000000000
            acc = (acc << 31) % 0x10000000000000000 | (acc >> 33)
            acc *= 0x9E3779B185EBCA87

        acc += len(v) ^ 2870177450013471926  # 0xC2B2AE3D27D4EB4F ^ 3527539
        acc %= 0x10000000000000000

        if acc == 0xFFFFFFFFFFFFFFFF:
            return 1546275796
        elif acc > 0x7FFFFFFFFFFFFFFF:  # (1 << 63) - 1 = largest positive number
            return acc - 0x10000000000000000
        return acc


class Morgan:
    __slots__ = ()

    @cached_property
    def atoms_order(self) -> Dict[int, int]:
        """
        Morgan like algorithm for graph nodes ordering

        :return: dict of atom-order pairs
        """
        atoms = self._atoms
        if not atoms:  # for empty containers
            return {}
        elif len(atoms) == 1:  # optimize single atom containers
            return dict.fromkeys(atoms, 1)
        return self._morgan({n: hash(a) for n, a in atoms.items()})

    def _morgan(self, weights: Dict[int, int]) -> Dict[int, int]:
        atoms = self._atoms
        bonds = self._bonds

        tries = len(atoms) - 1
        numb = len(set(weights.values()))
        stab = old_numb = 0

        for _ in range(tries):
            weights = {n: tuple_hash((weights[n],
                                      *(x for x in sorted((weights[m], int(b)) for m, b in ms.items()) for x in x)))
                       for n, ms in bonds.items()}
            old_numb, numb = numb, len(set(weights.values()))
            if numb == len(atoms):  # each atom now unique
                break
            elif numb == old_numb:  # not changed. molecules like benzene
                if min(Counter(weights.values()).items(), key=itemgetter(1))[1] > 1:
                    if stab == 3:
                        break
                elif stab >= 2:
                    break
                stab += 1
            elif stab:  # changed unique atoms number. reset stability check.
                stab = 0
        else:
            if numb < old_numb:
                warning('morgan. number of attempts exceeded. uniqueness has decreased.')

        return {n: i for i, (_, g) in enumerate(groupby(sorted(weights.items(), key=itemgetter(1)), key=itemgetter(1)),
                                                start=1) for n, _ in g}


__all__ = ['Morgan']



# -*- coding: utf-8 -*-
#
#  Copyright 2019, 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019, 2020 Dinar Batyrshin <batyrshin-dinar@mail.ru>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from importlib.util import find_spec
from itertools import combinations
from math import sqrt, pi, atan2, cos, sin
from random import uniform

if find_spec('numpy') and find_spec('numba'):  # try to load numba jit
    from numpy import array, zeros, uint16, zeros_like, empty
    from numba import b1, njit, f8, u2
else:
    def njit(*args, **kwargs):
        def wrapper(f):
            return f
        return wrapper

    class NumbaType:
        def __getitem__(self, item):
            return self

        def __call__(self, *args, **kwargs):
            return self

    b1 = f8 = u2 = NumbaType()


@njit(f8[:, :](f8[:, :], f8, f8, f8), cache=True)
def repulsive_force(xyz, coef, cut, inf):
    forces = zeros_like(xyz)
    n = len(xyz)
    inf **= 2
    cut **= 2

    for i in range(n - 1):
        xi, yi, zi = xyz[i]
        fxi, fyi, fzi = forces[i]
        for j in range(i + 1, n):
            xj, yj, zj = xyz[j]
            dx, dy, dz = xi - xj, yi - yj, zi - zj

            distance = dx ** 2 + dy ** 2 + dz ** 2
            if distance < cut:
                c = coef / cut
            elif distance > inf:
                c = 0.
            else:
                c = coef / distance

            # calculate repulsive force for each dimension
            dx, dy, dz = dx * c, dy * c, dz * c

            fxi += dx
            fyi += dy
            fzi += dz

            fxj, fyj, fzj = forces[j]
            forces[j] = fxj - dx, fyj - dy, fzj - dz
        forces[i] = fxi, fyi, fzi
    return forces


@njit(f8[:, :](f8[:, :], u2[:, :], f8, f8, f8), cache=True)
def straight_repulsive(xyz, straights, coef, cut, inf):
    forces = zeros_like(xyz)
    inf **= 2
    cut **= 2

    for n in range(len(straights)):
        i, j = straights[n]
        xi, yi, zi = xyz[i]
        xj, yj, zj = xyz[j]
        dx, dy, dz = xi - xj, yi - yj, zi - zj

        distance = dx ** 2 + dy ** 2 + dz ** 2
        if distance < cut:
            c = coef / cut
        elif distance > inf:
            c = 0.
        else:
            c = coef / distance

        dx, dy, dz = dx * c, dy * c, dz * c
        fxi, fyi, fzi = forces[i]
        fxj, fyj, fzj = forces[j]
        forces[i] = fxi + dx, fyi + dy, fzi + dz
        forces[j] = fxj - dx, fyj - dy, fzj - dz
    return forces


@njit(f8[:, :](f8[:, :], u2[:, :], f8[:, :]), cache=True)
def spring_force(xyz, springs, springs_distances):
    forces = zeros_like(xyz)
    for i in range(len(springs)):
        n, m = springs[i]
        r, stiff = springs_distances[i]
        nx, ny, nz = xyz[n]
        mx, my, mz = xyz[m]

        dx, dy, dz = nx - mx, ny - my, nz - mz
        distance = sqrt(dx ** 2 + dy ** 2 + dz ** 2)
        f = stiff * (distance - r)

        if -.001 < distance < .001:
            fdx, fdy, fdz = f, .0, .0
        else:
            f /= distance
            fdx, fdy, fdz = f * dx, f * dy, f * dz

        fxi, fyi, fzi = forces[n]
        fxj, fyj, fzj = forces[m]
        forces[n] = fxi - fdx, fyi - fdy, fzi - fdz
        forces[m] = fxj + fdx, fyj + fdy, fzj + fdz

    return forces


@njit(f8[:, :](f8[:, :], f8[:, :], f8), cache=True)
def flattening(forces, xyz, fc):
    n = len(forces)
    ff = zeros_like(xyz)
    for i in range(n):
        x, y, z = xyz[i]
        fx, fy, fz = forces[i]
        if -fc <= z <= fc:
            fz = -z
        else:
            if fz > 0 < z:
                fz -= fc
            elif fz < 0 > z:
                fz += fc
        ff[i] = fx, fy, fz
    return ff


@njit(f8[:, :](f8[:, :], f8), cache=True)
def cutoff(forces, cut):
    for n in range(len(forces)):
        dx, dy, dz = forces[n]
        distance = sqrt(dx ** 2 + dy ** 2 + dz ** 2)
        if distance > cut:
            f = cut / distance
            forces[n] = f * dx, f * dy, f * dz
    return forces


@njit((f8[:, :], b1[:, :], u2), cache=True)
def calculate_center(xyz, sssr_matrix, start):
    for n in range(len(sssr_matrix)):
        k = 0
        line = sssr_matrix[n]
        center_x, center_y, center_z = .0, .0, .0
        for i, b in enumerate(line):
            if b:
                k += 1
                x, y, z = xyz[i]
                center_x += x
                center_y += y
                center_z += z
        xyz[n + start] = center_x / k, center_y / k, center_z / k


@njit(f8[:, :](f8[:, :], u2[:, :], u2[:, :], f8[:, :], b1[:, :], u2), cache=True)
def steps(xyz, springs, straights, distances_stiffness, sssr_matrix, start_centers):
    # step 1
    for _ in range(2000):
        r_forces = repulsive_force(xyz, .05, .2, 10)
        s_forces = spring_force(xyz, springs, distances_stiffness)
        forces = r_forces + s_forces
        forces = cutoff(forces, .3)
        xyz = forces + xyz
        calculate_center(xyz, sssr_matrix, start_centers)

    # step 2
    for _ in range(1000):
        r_forces = repulsive_force(xyz, .02, .3, 3)
        s_forces = spring_force(xyz, springs, distances_stiffness)
        forces = r_forces + s_forces
        forces = flattening(forces, xyz, .1)
        forces = cutoff(forces, .3)
        xyz = forces + xyz
        calculate_center(xyz, sssr_matrix, start_centers)

    # step 3
    for _ in range(1000):
        r_forces = repulsive_force(xyz, .005, .4, 1.17) + straight_repulsive(xyz, straights, .05, .4, 2.)
        s_forces = spring_force(xyz, springs, distances_stiffness)
        forces = r_forces + s_forces
        forces = flattening(forces, xyz, .1)
        forces = cutoff(forces, .3)
        xyz = forces + xyz
        calculate_center(xyz, sssr_matrix, start_centers)

    return xyz


@njit(f8[:](f8[:, :], u2[:, :], u2), cache=True)
def get_angles(xyz, springs, bonds_count):
    angles = zeros(bonds_count)
    for i in range(bonds_count):
        n, m = springs[i]
        nx, ny = xyz[n, :2]
        mx, my = xyz[m, :2]
        nm = atan2(mx - nx, my - ny)
        if nm < 0:
            nm += pi
        angles[i] = nm
    return angles


@njit(f8[:, :](f8[:, :], u2, f8, f8), {'p': u2}, cache=True)
def rotate(xyz, atoms_count, shift_x, angle):
    cos_rad = cos(angle)
    sin_rad = sin(angle)
    xy = zeros((atoms_count, 2))

    dx, dy = xyz[0, :2]
    for p in range(1, atoms_count):
        px, py = xyz[p, :2]
        px, py = px - dx, py - dy
        xy[p] = cos_rad * px - sin_rad * py, sin_rad * px + cos_rad * py

    shift_y = xy[:, 1].mean()
    shift_r = shift_x - xy[:, 0].min()
    for p in range(atoms_count):
        px, py = xy[p]
        xy[p] = px + shift_r, py - shift_y
    return xy


class Calculate2D:
    __slots__ = ()

    def __prepare(self, component, randomize, c_stiff, r_stiff):
        atoms = self._atoms
        bonds = self._bonds
        plane = self._plane
        rings = [x for x in self.sssr if all(y in component for y in x)]

        mapping = {}
        springs = []
        straights = []
        xyz_matrix = []
        distances_stiffness = []
        dd = defaultdict(dict)

        # for cycles of 4 and 6 atoms
        c6 = []
        ac4 = set()
        ac6 = set()
        cycles_atoms = set()
        for ccl in rings:
            k = len(ccl)
            cycles_atoms.update(set(ccl))
            if k == 4:
                for a in ccl:
                    ac4.add(a)
            elif k == 6:
                c6.append(ccl)
                for a in ccl:
                    ac6.add(a)

        # create matrix of coordinates
        cube = len(component)
        if randomize:
            for i, n in enumerate(component):
                mapping[n] = i
                xyz_matrix.append([uniform(-cube, cube), uniform(-cube, cube), uniform(-cube, cube)])
        else:
            for i, n in enumerate(component):
                mapping[n] = i
                xyz_matrix.append([*plane[n], .1])

        # create matrix of connecting, springs and springs distances
        for n, m_bond in sorted(bonds.items(), reverse=True, key=lambda x: len(x[1])):
            if n not in component:
                continue
            for m, b in m_bond.items():
                if n not in dd[m]:
                    b = b.order
                    if b == 8 or len(m_bond) > 4:
                        dd[n][m] = dd[m][n] = True
                    else:
                        dd[n][m] = dd[m][n] = False

            # for angle = 180'
            if len(m_bond) == 2:
                (a1, bond1), (a2, bond2) = m_bond.items()
                order1, order2 = bond1.order, bond2.order
                if order1 == order2 == 2 or order1 == 3 or order2 == 3:
                    straights.append((mapping[a1], mapping[a2]))

        for n, m_bond in dd.items():
            i = mapping[n]
            for m, dist in m_bond.items():
                j = mapping[m]
                if (j, i) not in springs:
                    springs.append((i, j))
                    if n in cycles_atoms and m in cycles_atoms:
                        stiff = c_stiff
                    else:
                        stiff = r_stiff
                    distances_stiffness.append([1.32, stiff] if dist else [.825, stiff])

        bonds_count = len(distances_stiffness)
        # add virtual atoms and complement matrices of bonds and coordinates
        end = cube
        for n, m_bond in bonds.items():
            if n not in component:
                continue
            if len(m_bond) == 2:
                atom = atoms[n]
                if atom.atomic_number in {5, 6, 7, 8, 14, 15, 16, 17, 33, 34, 35, 52, 53, 85}:
                    (a1, bond1), (a2, bond2) = m_bond.items()
                    if self._is_angle(bond1, bond2):
                        mapping[-n] = end
                        dd[n][-n] = dd[-n][n] = False
                        if randomize:
                            xyz_matrix.append([uniform(-cube, cube), uniform(-cube, cube), uniform(-cube, cube)])
                        else:
                            xyz_matrix.append([*plane[n], -.1])
                        springs.append((mapping[n], end))
                        distances_stiffness.append([.825, r_stiff])
                        end += 1

        # add cycle centers
        start_centers = end
        sssr_matrix = []
        for ring in rings:
            k = len(ring)
            if k >= 11:  # skip big rings
                continue
            mapping[-end] = end
            sme = [False] * start_centers
            sssr_matrix.append(sme)
            for a, m in zip(ring, [mapping[x] for x in ring]):
                ns = bonds[a]
                sme[m] = True
                if len(ns) == 3:
                    for n in ns:
                        if n not in ring:
                            springs.append((mapping[n], end))
                            distances_stiffness.append([.825 + c_long[k], r_stiff])
            xyz_matrix.append([.0, .0, .0])
            end += 1
        xyz_matrix = array(xyz_matrix)
        if sssr_matrix:
            sssr_matrix = array(sssr_matrix, dtype=bool)
        else:
            sssr_matrix = zeros((0, start_centers), dtype=bool)
        calculate_center(xyz_matrix, sssr_matrix, start_centers)

        # add springs between cycles
        ini = start_centers
        for i, ring in enumerate(rings):
            ln = len(ring)
            if ln >= 11:
                continue
            clc = ini
            for ccl in rings[i+1:]:
                k = len(ccl)
                if k >= 11:
                    continue
                clc += 1
                if not set(ring).isdisjoint(set(ccl)):
                    springs.append((ini, clc))
                    distances_stiffness.append([c_short[ln] + c_short[k], c_stiff])
            ini += 1

        # add springs for cycles of six atoms
        for cycle in c6:
            for i, j in zip(cycle[:3], cycle[3:]):
                springs.append((mapping[i], mapping[j]))
                distances_stiffness.append([1.65, c_stiff])
        # add virtual bonds and complement matrices of bonds and coordinates
        for n, m_bond in dd.items():
            if len(m_bond) == 3:
                if n in cycles_atoms:
                    stiff = c_stiff
                else:
                    stiff = r_stiff
                for m1, m2 in combinations(m_bond, 2):
                    if m2 not in dd[m1]:
                        i, j = mapping[m1], mapping[m2]
                        if (j, i) not in springs or (i, j) not in springs:
                            springs.append((i, j))
                            long1, long2 = m_bond[m1], m_bond[m2]
                            if m1 in ac4 and m2 in ac4:
                                distances_stiffness.append([1.17, stiff])
                            elif not long1 and not long2:
                                distances_stiffness.append([1.43, stiff])
                            elif long1 or long2:
                                distances_stiffness.append([1.87, stiff])
                            else:
                                distances_stiffness.append([2.29, stiff])

        if not straights:
            straights = empty(shape=(0, 2), dtype=uint16)
        else:
            straights = array(straights, dtype=uint16)
        return xyz_matrix, array(springs, dtype=uint16), straights, array(distances_stiffness), cube, bonds_count, \
            sssr_matrix, start_centers

    def _is_angle(self, bond1, bond2):
        pass

    @staticmethod
    def __finish_xyz(xyz, springs, atoms_count, bonds_count, shift_x):
        angles = get_angles(xyz, springs, bonds_count)

        clusters = {}
        bonds = list(range(bonds_count))
        d = pi / 60
        while bonds:
            n = bonds.pop(0)
            nm = angles[n]
            values = [nm]
            for m in range(1, len(bonds)):
                m = angles[-m]
                if nm - d <= m <= nm + d:
                    values.append(m)
            clusters[n] = values

        angles = max((x for x in clusters.items()), key=lambda x: len(x[1]))[1]
        angle = sum(angles) / len(angles)
        xy = rotate(xyz, atoms_count, shift_x, angle)
        return xy, xy[:, 0].max() + .825

    def clean2d(self, *, randomize=False, cycle_stiff=.1, bond_stiff=.05):
        """
        Calculate 2d layout of graph.

        Idea from article:
        Frączek, T. (2016). Simulation-Based Algorithm for Two-Dimensional Chemical Structure Diagram Generation of
        Complex Molecules and Ligand–Protein Interactions. Journal of Chemical Information and Modeling, 56(12),
        2320–2335. doi:10.1021/acs.jcim.6b00391 (https://doi.org/10.1021/acs.jcim.6b00391)

        :param randomize: if True generating random coordinates for molecule
        :param cycle_stiff: stiffness for springs in cycles
        :param bond_stiff: stiffness for other springs
        """
        plane = self._plane

        shift_x = .0
        for component in self.connected_components:
            if len(component) == 1:
                plane[component[0]] = (shift_x, .0)
                shift_x += .825
            elif len(component) == 2:
                plane[component[0]] = (shift_x, .0)
                plane[component[1]] = (shift_x, .825)
                shift_x += .825
            else:
                if not randomize and all(-.0001 < x[0] < .0001 and -.0001 < x[1] < .0001 for x in plane.values()):
                    randomize = True
                xyz, springs, straights, distances_stiffness, atoms_count, bonds_count, sssr_matrix, start_centers = \
                    self.__prepare(component, randomize, cycle_stiff, bond_stiff)
                xyz = steps(xyz, springs, straights, distances_stiffness, sssr_matrix, start_centers)
                xy, shift_x = self.__finish_xyz(xyz, springs, atoms_count, bonds_count, shift_x)
                for i, n in enumerate(component):
                    plane[n] = tuple(xy[i])
        self.__dict__.pop('__cached_method__repr_svg_', None)


class Calculate2DMolecule(Calculate2D):
    __slots__ = ()

    def _is_angle(self, bond1, bond2):
        order1, order2 = bond1.order, bond2.order
        return not (order1 == order2 == 2 or order1 == 3 or order2 == 3 or order1 == 8 or order2 == 8)


class Calculate2DCGR(Calculate2D):
    __slots__ = ()

    @staticmethod
    def __primary(bond):
        w_order, p_order = bond.order, bond.p_order
        if w_order is None:
            order = p_order
        elif p_order is None:
            order = w_order
        else:
            if w_order == 8:
                order = p_order
            elif p_order == 8:
                order = w_order
            else:
                if w_order > p_order:
                    order = w_order
                else:
                    order = p_order
        return order

    def _is_angle(self, bond1, bond2):
        order1 = self.__primary(bond1)
        order2 = self.__primary(bond2)
        return not (order1 == order2 == 2 or order1 == 3 or order2 == 3 or order1 == 8 or order2 == 8)


if find_spec('numpy'):  # load numpy not yet loaded with numba
    if not find_spec('numba'):
        from numpy import array, zeros, uint16, zeros_like, empty
else:  # disable clean2d support
    class Calculate2DMolecule:
        __slots__ = ()

        def clean2d(self, **kwargs):
            raise NotImplemented('numpy required for clean2d')


    class Calculate2DCGR:
        __slots__ = ()

        def clean2d(self, **kwargs):
            raise NotImplemented('numpy required for clean2d')

c_long = {3: 1.65 * cos(2 * pi / 3), 4: 1.65 * cos(pi / 2), 5: 1.65 * cos(2 * pi / 5), 6: 1.65 * cos(pi / 3),
          7: 1.65 * cos(2 * pi / 7), 8: 1.65 * cos(pi / 4), 9: 1.65 * cos(2 * pi / 9), 10: 1.65 * cos(pi / 5)}

c_short = {3: c_long[3] * cos(pi / 3), 4: c_long[4] * cos(pi / 4), 5: c_long[5] * cos(pi / 5),
           6: c_long[6] * cos(pi / 6), 7: c_long[7] * cos(pi / 7), 8: c_long[8] * cos(pi / 8),
           9: c_long[9] * cos(pi / 9), 10: c_long[10] * cos(pi / 10)}

__all__ = ['Calculate2DMolecule', 'Calculate2DCGR']



# -*- coding: utf-8 -*-
#
#  Copyright 2019 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Alexander Nikanshin <17071996sasha@gmail.com>
#  Copyright 2019 Tagir Akhmetshin <tagirshin@gmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import FrozenDict
from .element import Element
from .groups import GroupV
from .periods import PeriodIV, PeriodV, PeriodVI, PeriodVII


class V(Element, PeriodIV, GroupV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 23

    @property
    def isotopes_distribution(self):
        return FrozenDict({50: 0.0025, 51: 0.9975})

    @property
    def isotopes_masses(self):
        return FrozenDict({50: 49.947163, 51: 50.943964})

    @property
    def _common_valences(self):
        return 0, 2

    @property
    def _valences_exceptions(self):
        return ((2, False, 0, ()), (3, False, 0, ()), (2, False, 0, ((2, 'O'),)),
                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((2, 'O'), (1, 'O'))),
                (0, False, 0, ((2, 'O'), (1, 'Cl'))),

                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'), (1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((2, 'O'), (2, 'O'))),
                (0, False, 0, ((2, 'O'), (1, 'Cl'), (1, 'Cl'))),

                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((2, 'O'), (2, 'O'), (1, 'O'))),
                (0, False, 0, ((2, 'S'), (2, 'S'), (1, 'S'))),
                (0, False, 0, ((2, 'O'), (1, 'O'), (1, 'O'), (1, 'O'))),
                (0, False, 0, ((2, 'O'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((2, 'O'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))))

    @property
    def atomic_radius(self):
        return 1.71


class Nb(Element, PeriodV, GroupV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 41

    @property
    def isotopes_distribution(self):
        return FrozenDict({93: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({93: 92.906378})

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ((0, False, 0, ((2, 'O'),)),
                (0, False, 0, ((2, 'O'), (1, 'O'))),
                (0, False, 0, ((3, 'N'),)),

                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'), (1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((2, 'O'), (2, 'O'))),
                (0, False, 0, ((2, 'S'), (2, 'S'))),

                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((2, 'O'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((2, 'O'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((2, 'O'), (2, 'O'), (1, 'O'))),
                (0, False, 0, ((2, 'O'), (1, 'O'), (1, 'O'), (1, 'O'))))

    @property
    def atomic_radius(self):
        return 1.98


class Ta(Element, PeriodVI, GroupV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 73

    @property
    def isotopes_distribution(self):
        return FrozenDict({180: 0.00012, 181: 0.99988})

    @property
    def isotopes_masses(self):
        return FrozenDict({180: 179.947466, 181: 180.947996})

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ((0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((2, 'O'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((2, 'O'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((2, 'O'), (2, 'O'), (1, 'O'))),
                (0, False, 0, ((2, 'O'), (1, 'O'), (1, 'O'), (1, 'O'))))

    @property
    def atomic_radius(self):
        return 2.0


class Db(Element, PeriodVII, GroupV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 105

    @property
    def isotopes_distribution(self):
        return FrozenDict({268: 1.0})

    @property
    def isotopes_masses(self):
        return FrozenDict({268: 268.125676})

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ()

    @property
    def atomic_radius(self):
        return 2.0  # unknown, taken radius of previous element in group


__all__ = ['V', 'Nb', 'Ta', 'Db']



# -*- coding: utf-8 -*-
#
#  Copyright 2014-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict, namedtuple
from itertools import count
from ..containers import CGRContainer, MoleculeContainer, QueryContainer, ReactionContainer
from ..containers.bonds import Bond, DynamicBond
from ..exceptions import AtomNotFound, MappingError
from ..periodictable import DynamicElement, Element, QueryElement


common_isotopes = {'H': 1, 'He': 4, 'Li': 7, 'Be': 9, 'B': 11, 'C': 12, 'N': 14, 'O': 16, 'F': 19, 'Ne': 20, 'Na': 23,
                   'Mg': 24, 'Al': 27, 'Si': 28, 'P': 31, 'S': 32, 'Cl': 35, 'Ar': 40, 'K': 39, 'Ca': 40, 'Sc': 45,
                   'Ti': 48, 'V': 51, 'Cr': 52, 'Mn': 55, 'Fe': 56, 'Co': 59, 'Ni': 59, 'Cu': 64, 'Zn': 65, 'Ga': 70,
                   'Ge': 73, 'As': 75, 'Se': 79, 'Br': 80, 'Kr': 84, 'Rb': 85, 'Sr': 88, 'Y': 89, 'Zr': 91, 'Nb': 93,
                   'Mo': 96, 'Tc': 98, 'Ru': 101, 'Rh': 103, 'Pd': 106, 'Ag': 108, 'Cd': 112, 'In': 115, 'Sn': 119,
                   'Sb': 122, 'Te': 128, 'I': 127, 'Xe': 131, 'Cs': 133, 'Ba': 137, 'La': 139, 'Ce': 140, 'Pr': 141,
                   'Nd': 144, 'Pm': 145, 'Sm': 150, 'Eu': 152, 'Gd': 157, 'Tb': 159, 'Dy': 163, 'Ho': 165, 'Er': 167,
                   'Tm': 169, 'Yb': 173, 'Lu': 175, 'Hf': 178, 'Ta': 181, 'W': 184, 'Re': 186, 'Os': 190, 'Ir': 192,
                   'Pt': 195, 'Au': 197, 'Hg': 201, 'Tl': 204, 'Pb': 207, 'Bi': 209, 'Po': 209, 'At': 210, 'Rn': 222,
                   'Fr': 223, 'Ra': 226, 'Ac': 227, 'Th': 232, 'Pa': 231, 'U': 238, 'Np': 237, 'Pu': 244, 'Am': 243,
                   'Cm': 247, 'Bk': 247, 'Cf': 251, 'Es': 252, 'Fm': 257, 'Md': 258, 'No': 259, 'Lr': 260, 'Rf': 261,
                   'Db': 270, 'Sg': 269, 'Bh': 270, 'Hs': 270, 'Mt': 278, 'Ds': 281, 'Rg': 281, 'Cn': 285, 'Nh': 278,
                   'Fl': 289, 'Mc': 289, 'Lv': 293, 'Ts': 297, 'Og': 294}
elements_set = set(common_isotopes)
elements_list = list(common_isotopes)


parse_error = namedtuple('ParseError', ('number', 'position', 'log'))


class CGRRead:
    def __init__(self, remap=True, ignore=False, store_log=False):
        self.__remap = remap
        self._ignore = ignore
        self._store_log = store_log
        self._log_buffer = []

    def _info(self, msg):
        self._log_buffer.append(msg)

    def _flush_log(self):
        self._log_buffer.clear()

    def _format_log(self):
        return '\n'.join(self._log_buffer)

    def _convert_reaction(self, reaction):
        if not (reaction['reactants'] or reaction['products'] or reaction['reagents']):
            raise ValueError('empty reaction')
        maps = {'reactants': [], 'products': [], 'reagents': []}
        for i, tmp in maps.items():
            for molecule in reaction[i]:
                used = set()
                for atom in molecule['atoms']:
                    m = atom['mapping']
                    if m:
                        if m in used:
                            if not self._ignore:
                                raise MappingError('mapping in molecules should be unique')
                            self._info(f'non-unique mapping in molecule: {m}')
                        else:
                            used.add(m)
                    tmp.append(m)

        length = count(max(max(maps['products'], default=0), max(maps['reactants'], default=0),
                           max(maps['reagents'], default=0)) + 1)

        # map unmapped atoms.
        for i, tmp in maps.items():
            used = set()
            maps[i] = remap = []
            for m in tmp:
                if not m:
                    remap.append(next(length))
                elif m in used:
                    if not self._ignore:
                        raise MappingError('mapping in reagents or products or reactants should be unique')
                    # force remap non unique atoms in molecules.
                    remap.append(next(length))
                    self._info(f'mapping in {i} changed from {m} to {remap[-1]}')
                else:
                    remap.append(m)
                    used.add(m)

        if maps['reagents']:
            tmp = (set(maps['reactants']) | set(maps['products'])) & set(maps['reagents'])
            if tmp:
                e = f'reagents has map intersection with reactants or products: {tmp}'
                if not self._ignore:
                    raise MappingError(e)
                self._info(e)
                maps['reagents'] = [x if x not in tmp else next(length) for x in maps['reagents']]

        # find breaks in map. e.g. 1,2,5,6. 3,4 - skipped
        if self.__remap:
            lose = sorted(set(range(1, next(length))) - set(maps['reactants']) - set(maps['products']) -
                          set(maps['reagents']), reverse=True)
            if lose:
                for i, tmp in maps.items():
                    if not tmp:
                        continue
                    for j in lose:
                        maps[i] = tmp = [x if x < j else x - 1 for x in tmp]

        rc = {'reactants': [], 'products': [], 'reagents': []}
        for i, tmp in maps.items():
            shift = 0
            for j in reaction[i]:
                atom_len = len(j['atoms'])
                remapped = {x: y for x, y in enumerate(tmp[shift: atom_len + shift])}
                shift += atom_len
                g = self.__prepare_structure(j, remapped)
                rc[i].append(g)
        return ReactionContainer(meta=reaction['meta'], name=reaction.get('title'), **rc)

    def _convert_structure(self, molecule):
        if self.__remap:
            remapped = {n: k for n, k in enumerate(range(1, len(molecule['atoms']) + 1))}
        else:
            length = count(max(x['mapping'] for x in molecule['atoms']) + 1)
            remapped, used = {}, set()
            for n, atom in enumerate(molecule['atoms']):
                m = atom['mapping']
                if not m:
                    remapped[n] = next(length)
                elif m in used:
                    if not self._ignore:
                        raise MappingError('mapping in molecules should be unique')
                    remapped[n] = next(length)
                    self._info(f'mapping in molecule changed from {m} to {remapped[n]}')
                else:
                    remapped[n] = m
                    used.add(m)

        g = self.__prepare_structure(molecule, remapped)
        g.meta.update(molecule['meta'])
        return g

    @staticmethod
    def _convert_molecule(molecule, mapping):
        g = object.__new__(MoleculeContainer)
        pm = {}
        atoms = {}
        plane = {}
        charges = {}
        radicals = {}
        bonds = {}
        for n, atom in enumerate(molecule['atoms']):
            n = mapping[n]
            atoms[n] = Element.from_symbol(atom['element'])(atom['isotope'])
            bonds[n] = {}

            charges[n] = g._validate_charge(atom['charge'])
            radicals[n] = atom['is_radical']
            plane[n] = (atom['x'], atom['y'])
            pm[n] = atom['mapping']
        for n, m, b in molecule['bonds']:
            n, m = mapping[n], mapping[m]
            if n == m:
                raise ValueError('atom loops impossible')
            if n not in bonds or m not in bonds:
                raise AtomNotFound('atoms not found')
            if n in bonds[m]:
                raise ValueError('atoms already bonded')
            bonds[n][m] = bonds[m][n] = Bond(b)
        if any(a['z'] for a in molecule['atoms']):
            conformers = [{mapping[n]: (a['x'], a['y'], a['z']) for n, a in enumerate(molecule['atoms'])}]
        else:
            conformers = []
        g.__setstate__({'atoms': atoms, 'bonds': bonds, 'meta': {}, 'plane': plane, 'parsed_mapping': pm,
                        'charges': charges, 'radicals': radicals, 'name': '', 'conformers': conformers,
                        'atoms_stereo': {}, 'allenes_stereo': {}, 'cis_trans_stereo': {}})
        return g

    @staticmethod
    def _convert_cgr(molecule, mapping):
        atoms = molecule['atoms']
        bonds = defaultdict(dict)

        for nm, _type, value in molecule['cgr']:
            if _type == 'radical':
                atoms[nm]['p_is_radical'] = not atoms[nm]['is_radical']
            elif _type == 'charge':
                atoms[nm]['p_charge'] = atoms[nm]['charge'] + value
            else:
                n, m = nm
                bonds[n][m] = bonds[m][n] = DynamicBond(*value)

        g = object.__new__(CGRContainer)
        pm = {}
        g_atoms = {}
        plane = {}
        charges = {}
        radicals = {}
        p_charges = {}
        p_radicals = {}
        g_bonds = {}
        for n, atom in enumerate(atoms):
            n = mapping[n]
            g_atoms[n] = DynamicElement.from_symbol(atom['element'])(atom['isotope'])
            g_bonds[n] = {}
            charges[n] = g._validate_charge(atom['charge'])
            radicals[n] = atom['is_radical']
            p_charges[n] = g._validate_charge(atom.get('p_charge', atom['charge']))
            p_radicals[n] = atom.get('p_is_radical', atom['is_radical'])
            plane[n] = (atom['x'], atom['y'])
            pm[n] = atom['mapping']
        for n, m, b in molecule['bonds']:
            if m in bonds[n]:
                if b != 8:
                    raise ValueError('CGR spec invalid')
                b = bonds[n][m]
            else:
                b = DynamicBond(b, b)
            n, m = mapping[n], mapping[m]
            if n == m:
                raise ValueError('atom loops impossible')
            if n not in g_bonds or m not in g_bonds:
                raise AtomNotFound('atoms not found')
            if n in g_bonds[m]:
                raise ValueError('atoms already bonded')
            g_bonds[n][m] = g_bonds[m][n] = b
        g.__setstate__({'atoms': g_atoms, 'bonds': g_bonds, 'meta': {}, 'plane': plane, 'parsed_mapping': pm,
                        'charges': charges, 'radicals': radicals, 'name': '', 'p_charges': p_charges,
                        'p_radicals': p_radicals})
        return g

    @staticmethod
    def _convert_query(molecule, mapping):
        atoms = molecule['atoms']
        for n, _type, value in molecule['query']:
            atoms[n][_type] = value

        g = QueryContainer()
        pm = g._parsed_mapping
        for n, atom in enumerate(atoms):
            n = g.add_atom(QueryElement.from_symbol(atom['element'])(atom['isotope']), mapping[n],
                           charge=atom['charge'], is_radical=atom['is_radical'],
                           neighbors=atom.get('neighbors'), hybridization=atom.get('hybridization'),
                           xy=(atom['x'], atom['y']))
            pm[n] = atom['mapping']
        for n, m, b in molecule['bonds']:
            g.add_bond(mapping[n], mapping[m], b)
        return g

    def __prepare_structure(self, molecule, mapping):
        if 'query' in molecule:
            if 'cgr' in molecule:
                raise ValueError('QueryCGR parsing not supported')
            g = self._convert_query(molecule, mapping)
        elif 'cgr' in molecule:
            g = self._convert_cgr(molecule, mapping)
        else:
            g = self._convert_molecule(molecule, mapping)

        if 'title' in molecule:
            g.name = molecule['title']
        return g



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from typing import List, Union, Tuple, Dict, Optional
from . import cgr_query as query, molecule  # cyclic imports resolve
from .bonds import Bond, DynamicBond
from .common import Graph
from ..algorithms.calculate2d import Calculate2DCGR
from ..algorithms.components import CGRComponents
from ..algorithms.depict import DepictCGR
from ..algorithms.smiles import CGRSmiles
from ..algorithms.x3dom import X3domCGR
from ..exceptions import MappingError
from ..periodictable import DynamicElement, Element, DynamicQueryElement


class CGRContainer(Graph, CGRSmiles, CGRComponents, DepictCGR, Calculate2DCGR, X3domCGR):
    __slots__ = ('_conformers', '_p_charges', '_p_radicals', '_hybridizations', '_p_hybridizations')

    def __init__(self):
        self._conformers: List[Dict[int, Tuple[float, float, float]]] = []
        self._p_charges: Dict[int, int] = {}
        self._p_radicals: Dict[int, bool] = {}
        self._hybridizations: Dict[int, int] = {}
        self._p_hybridizations: Dict[int, int] = {}
        super().__init__()

    def add_atom(self, atom: Union[DynamicElement, Element, int, str], *args, p_charge: int = 0,
                 p_is_radical: bool = False, **kwargs):
        p_charge = self._validate_charge(p_charge)
        p_is_radical = self._validate_radical(p_is_radical)

        if not isinstance(atom, DynamicElement):
            if isinstance(atom, Element):
                atom = DynamicElement.from_atomic_number(atom.atomic_number)(atom.isotope)
            elif isinstance(atom, str):
                atom = DynamicElement.from_symbol(atom)()
            elif isinstance(atom, int):
                atom = DynamicElement.from_atomic_number(atom)()
            else:
                raise TypeError('DynamicElement object expected')

        _map = super().add_atom(atom, *args, **kwargs)
        self._p_charges[_map] = p_charge
        self._p_radicals[_map] = p_is_radical
        self._hybridizations[_map] = 1
        self._p_hybridizations[_map] = 1
        self._conformers.clear()  # clean conformers. need full recalculation for new system
        return _map

    def add_bond(self, n, m, bond: Union[DynamicBond, Bond, int]):
        if isinstance(bond, DynamicBond):
            order = bond.order
            p_order = bond.p_order
        elif isinstance(bond, Bond):
            order = p_order = bond.order
            bond = object.__new__(DynamicBond)
            bond._DynamicBond__order = bond._DynamicBond__p_order = order
        else:
            order = p_order = bond
            bond = DynamicBond(order, order)

        super().add_bond(n, m, bond)
        self._conformers.clear()

        if order != 1 or p_order != 1:  # 1 is neutral.
            self._calc_hybridization(n)
            self._calc_hybridization(m)

    def delete_atom(self, n):
        old_bonds = self._bonds[n]  # save bonds
        super().delete_atom(n)

        del self._p_charges[n]
        del self._p_radicals[n]
        del self._hybridizations[n]
        del self._p_hybridizations[n]

        for m in old_bonds:
            self._calc_hybridization(m)
        self._conformers.clear()

    def delete_bond(self, n, m):
        super().delete_bond(n, m)
        self._conformers.clear()
        self._calc_hybridization(n)
        self._calc_hybridization(m)

    def remap(self, mapping, *, copy=False) -> 'CGRContainer':
        h = super().remap(mapping, copy=copy)
        mg = mapping.get
        spr = self._p_radicals
        sh = self._hybridizations
        sph = self._p_hybridizations

        if copy:
            hpc = h._p_charges
            hpr = h._p_radicals
            hh = h._hybridizations
            hc = h._conformers
            hph = h._p_hybridizations
        else:
            hpc = {}
            hpr = {}
            hh = {}
            hc = []
            hph = {}

        for n, c in self._p_charges.items():
            m = mg(n, n)
            hpc[m] = c
            hpr[m] = spr[n]
            hh[m] = sh[n]
            hph[m] = sph[n]

        hc.extend({mg(n, n): x for n, x in c.items()} for c in self._conformers)

        if copy:
            return h

        self._p_charges = hpc
        self._p_radicals = hpr
        self._hybridizations = hh
        self._conformers = hc
        self._p_hybridizations = hph
        return self

    def copy(self, **kwargs) -> 'CGRContainer':
        copy = super().copy(**kwargs)
        copy._hybridizations = self._hybridizations.copy()
        copy._conformers = [c.copy() for c in self._conformers]
        copy._p_hybridizations = self._p_hybridizations.copy()
        copy._p_radicals = self._p_radicals.copy()
        copy._p_charges = self._p_charges.copy()
        return copy

    def substructure(self, atoms, *, as_query: bool = False, **kwargs) -> Union['CGRContainer',
                                                                                'query.QueryCGRContainer']:
        """
        create substructure containing atoms from atoms list

        :param atoms: list of atoms numbers of substructure
        :param meta: if True metadata will be copied to substructure
        :param as_query: return Query object based on graph substructure
        """
        sub, atoms = super().substructure(atoms, query.QueryCGRContainer if as_query else self.__class__, **kwargs)
        sa = self._atoms
        spc = self._p_charges
        spr = self._p_radicals

        sub._p_charges = {n: spc[n] for n in atoms}
        sub._p_radicals = {n: spr[n] for n in atoms}

        if as_query:
            sub._atoms = ca = {}
            for n in atoms:
                atom = sa[n]
                atom = DynamicQueryElement.from_atomic_number(atom.atomic_number)(atom.isotope)
                ca[n] = atom
                atom._attach_to_graph(sub, n)

            sb = self._bonds
            sh = self._hybridizations
            sph = self._p_hybridizations
            sub._neighbors = {n: (sum(x.order is not None for x in sb[n].values()),) for n in atoms}
            sub._hybridizations = {n: (sh[n],) for n in atoms}
            sub._p_neighbors = {n: (sum(x.p_order is not None for x in sb[n].values()),) for n in atoms}
            sub._p_hybridizations = {n: (sph[n],) for n in atoms}
        else:
            sub._conformers = [{n: c[n] for n in atoms} for c in self._conformers]
            sub._atoms = ca = {}
            for n in atoms:
                atom = sa[n].copy()
                ca[n] = atom
                atom._attach_to_graph(sub, n)

            # recalculate query marks
                # recalculate query marks
            sub._hybridizations = {}
            sub._p_hybridizations = {}
            for n in sub._atoms:
                sub._calc_hybridization(n)
        return sub

    def union(self, other, **kwargs):
        if isinstance(other, CGRContainer):
            u, other = super().union(other, **kwargs)
            u._conformers.clear()

            u._p_charges.update(other._p_charges)
            u._p_radicals.update(other._p_radicals)
            u._hybridizations.update(other._hybridizations)
            u._p_hybridizations.update(other._p_hybridizations)

            ub = u._bonds
            for n in other._bonds:
                ub[n] = {}
            seen = set()
            for n, m_bond in other._bonds.items():
                seen.add(n)
                for m, bond in m_bond.items():
                    if m not in seen:
                        ub[n][m] = ub[m][n] = bond.copy()

            ua = u._atoms
            for n, atom in other._atoms.items():
                atom = atom.copy()
                ua[n] = atom
                atom._attach_to_graph(u, n)
            return u
        elif isinstance(other, molecule.MoleculeContainer):
            u, other = super().union(other, **kwargs)
            u._p_charges.update(other._charges)
            u._p_radicals.update(other._radicals)
            u._hybridizations.update(other._hybridizations)
            u._p_hybridizations.update(other._hybridizations)

            ub = u._bonds
            for n, m_bond in other._bonds.items():
                ub[n] = {m: DynamicBond(b.order, b.order) for m, b in m_bond.items()}

            ua = u._atoms
            for n, atom in other._atoms.items():
                atom = DynamicElement.from_atomic_number(atom.atomic_number)(atom.isotope)
                ua[n] = atom
                atom._attach_to_graph(u, n)
            return u
        elif isinstance(other, Graph):  # Query or CGRQuery
            return other.union(self, **kwargs)
        else:
            raise TypeError('Graph expected')

    def compose(self, other: Union['molecule.MoleculeContainer', 'CGRContainer']) -> 'CGRContainer':
        """
        compose 2 graphs to CGR

        :param other: Molecule or CGR Container
        :return: CGRContainer
        """
        sa = self._atoms
        sc = self._charges
        sr = self._radicals
        spc = self._p_charges
        spr = self._p_radicals
        sp = self._plane
        sb = self._bonds

        bonds = []
        adj: Dict[int, Dict[int, List[Optional[int]]]] = defaultdict(lambda: defaultdict(lambda: [None, None]))
        h = self.__class__()  # subclasses support
        atoms = h._atoms

        if isinstance(other, molecule.MoleculeContainer):
            oa = other._atoms
            oc = other._charges
            or_ = other._radicals
            op = other._plane
            ob = other._bonds
            common = sa.keys() & other

            for n in sa.keys() - common:  # cleavage atoms
                h.add_atom(sa[n].copy(), n, charge=sc[n], is_radical=sr[n], xy=sp[n], p_charge=spc[n],
                           p_is_radical=spr[n])
                for m, bond in sb[n].items():
                    if m not in atoms:
                        if m in common:  # bond to common atoms is broken bond
                            order = bond.order
                            if order:  # skip formed bond. None>X => None>None
                                bond = object.__new__(DynamicBond)
                                bond._DynamicBond__order, bond._DynamicBond__p_order = order, None
                                bonds.append((n, m, bond))
                        else:
                            bonds.append((n, m, bond))
            for n in other._atoms.keys() - common:  # coupling atoms
                h.add_atom(oa[n], n, charge=oc[n], is_radical=or_[n], xy=op[n], p_charge=oc[n], p_is_radical=or_[n])
                for m, bond in ob[n].items():
                    if m not in atoms:
                        if m in common:  # bond to common atoms is formed bond
                            order = bond.order
                            bond = object.__new__(DynamicBond)
                            bond._DynamicBond__order, bond._DynamicBond__p_order = None, order
                        bonds.append((n, m, bond))
            for n in common:
                an = adj[n]
                for m, bond in ob[n].items():
                    if m in common:
                        an[m][1] = bond.order
                for m, bond in sb[n].items():
                    if m in an or m in common and bond.order:
                        an[m][0] = bond.order
            for n in common:
                san = sa[n]
                if san.atomic_number != oa[n].atomic_number or san.isotope != oa[n].isotope:
                    raise MappingError(f'atoms with number {{{n}}} not equal')
                h.add_atom(san.copy(), n, charge=sc[n], is_radical=sr[n], xy=sp[n], p_charge=oc[n], p_is_radical=or_[n])
                for m, (o1, o2) in adj[n].items():
                    if m not in atoms:
                        bond = object.__new__(DynamicBond)
                        bond._DynamicBond__order, bond._DynamicBond__p_order = o1, o2
                        bonds.append((n, m, bond))
        elif isinstance(other, CGRContainer):
            oa = other._atoms
            oc = other._charges
            or_ = other._radicals
            opc = other._p_charges
            opr = other._p_radicals
            op = other._plane
            ob = other._bonds
            common = sa.keys() & other

            for n in sa.keys() - common:  # cleavage atoms
                h.add_atom(sa[n].copy(), n, charge=sc[n], is_radical=sr[n], xy=sp[n], p_charge=spc[n],
                           p_is_radical=spr[n])
                for m, bond in sb[n].items():
                    if m not in atoms:
                        if m in common:  # bond to common atoms is broken bond
                            order = bond.order
                            if order:  # skip formed bond. None>X => None>None
                                bond = object.__new__(DynamicBond)
                                bond._DynamicBond__order, bond._DynamicBond__p_order = order, None
                                bonds.append((n, m, bond))
                        else:
                            bonds.append((n, m, bond))
            for n in other._atoms.keys() - common:  # coupling atoms
                h.add_atom(oa[n].copy(), n, charge=oc[n], is_radical=or_[n], xy=op[n], p_charge=opc[n],
                           p_is_radical=opr[n])
                for m, bond in ob[n].items():
                    if m not in atoms:
                        if m in common:  # bond to common atoms is formed bond
                            order = bond.p_order
                            if order:  # skip broken bond. X>None => None>None
                                bond = object.__new__(DynamicBond)
                                bond._DynamicBond__order, bond._DynamicBond__p_order = None, order
                                bonds.append((n, m, bond))
                        else:
                            bonds.append((n, m, bond))
            for n in common:
                an = adj[n]
                for m, bond in sb[n].items():
                    if m in common and bond.order:  # skip formed bonds
                        an[m][0] = bond.order
                for m, bond in ob[n].items():
                    if m in an or m in common and bond.p_order:
                        # self has nm bond or other bond not broken
                        an[m][1] = bond.p_order
            for n in common:
                san = sa[n]
                if san.atomic_number != oa[n].atomic_number or san.isotope != oa[n].isotope:
                    raise MappingError(f'atoms with number {{{n}}} not equal')
                h.add_atom(san.copy(), n, charge=sc[n], is_radical=sr[n], xy=sp[n], p_charge=opc[n],
                           p_is_radical=opr[n])
                for m, (o1, o2) in adj[n].items():
                    if m not in atoms:
                        bond = object.__new__(DynamicBond)
                        bond._DynamicBond__order, bond._DynamicBond__p_order = o1, o2
                        bonds.append((n, m, bond))
        else:
            raise TypeError('MoleculeContainer or CGRContainer expected')

        for n, m, bond in bonds:
            h.add_bond(n, m, bond)
        return h

    def __xor__(self, other):
        """
        G ^ H is CGR generation
        """
        return self.compose(other)

    def get_mapping(self, other: 'CGRContainer', **kwargs):
        if isinstance(other, CGRContainer):
            return super().get_mapping(other, **kwargs)
        raise TypeError('CGRContainer expected')

    def get_mcs_mapping(self, other: 'CGRContainer', **kwargs):
        if isinstance(other, CGRContainer):
            return super().get_mcs_mapping(other, **kwargs)
        raise TypeError('CGRContainer expected')

    def decompose(self) -> Tuple['molecule.MoleculeContainer', 'molecule.MoleculeContainer']:
        """
        decompose CGR to pair of Molecules, which represents reactants and products state of reaction

        :return: tuple of two molecules
        """
        charges = self._charges
        p_charges = self._p_charges
        radicals = self._radicals
        p_radicals = self._p_radicals
        plane = self._plane

        reactants = molecule.MoleculeContainer()
        products = molecule.MoleculeContainer()

        for n, atom in self._atoms.items():
            atom = Element.from_atomic_number(atom.atomic_number)(atom.isotope)
            reactants.add_atom(atom, n, charge=charges[n], is_radical=radicals[n], xy=plane[n])
            products.add_atom(atom.copy(), n, charge=p_charges[n], is_radical=p_radicals[n], xy=plane[n])

        for n, m, bond in self.bonds():
            if bond.order:
                reactants.add_bond(n, m, bond.order)
            if bond.p_order:
                products.add_bond(n, m, bond.p_order)
        return reactants, products

    def __invert__(self):
        """
        decompose CGR
        """
        return self.decompose()

    def _calc_hybridization(self, n: int):
        hybridization = p_hybridization = 1
        for bond in self._bonds[n].values():
            order = bond.order
            p_order = bond.p_order
            if order and hybridization != 4:
                if order == 4:
                    hybridization = 4
                elif order == 3:
                    if hybridization != 3:
                        hybridization = 3
                elif order == 2:
                    if hybridization == 2:
                        hybridization = 3
                    elif hybridization == 1:
                        hybridization = 2
            if p_order and p_hybridization != 4:
                if p_order == 4:
                    p_hybridization = 4
                elif p_order == 3:
                    if p_hybridization != 3:
                        p_hybridization = 3
                elif p_order == 2:
                    if p_hybridization == 2:
                        p_hybridization = 3
                    elif p_hybridization == 1:
                        p_hybridization = 2
        self._hybridizations[n] = hybridization
        self._p_hybridizations[n] = p_hybridization

    def __getstate__(self):
        return {'conformers': self._conformers, 'p_charges': self._p_charges, 'p_radicals': self._p_radicals,
                **super().__getstate__()}

    def __setstate__(self, state):
        self._p_charges = state['p_charges']
        self._p_radicals = state['p_radicals']
        super().__setstate__(state)
        if 'conformers' in state:
            self._conformers = state['conformers']
        else:
            self._conformers = []  # < 4.0.23 compatibility

        # restore query marks
        self._hybridizations = {}
        self._p_hybridizations = {}
        for n in state['bonds']:
            self._calc_hybridization(n)


__all__ = ['CGRContainer']
